#
# Containers for time-series of 1d and 2d rectangular data arrays.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import sys
import array
import numpy as np
import myokit


# Readme file for DataBlock1d binary files
README_SAVE_1D = """
Myokit DataBlock1d Binary File
==============================
This zip file contains simulation data in the form of multiple time series.
Zero-dimensional time series, such as time or a global pace variable are
stored, as well as one-dimensional time series, such as the membrane potential
of a linear sequence of cells, as it varies over time.

This file has the following entries:

header_block1d.txt
------------------
A header file containing the following information (line by line):

  - nt     the number of points in time in each entry
  - nx     the width of each 1d block
  - dtype  the used datatype (either "d" or "f")
  - name   the names of all 0d entries, each on its own line
  - 1      the indication that the 1d entries are starting
  - name   the names of all 1d entries, each on its own line

data.bin
--------
A binary file containing the following data, in the data type specified by the
header, and little-endian:

  - The nt time values
  - All 0d entries
  - All 1d entries, reshaped using numpy order='C'

""".strip()

# Readme file for DataBlock2d binary files
README_SAVE_2D = """
Myokit DataBlock2d Binary File
==============================
This zip file contains simulation data in the form of multiple time series.
Zero-dimensional time series, such as time or a global pace variable are
stored, as well as two-dimensional time series, such as the membrane potential
of a 2d grid of cells, as it varies over time.

This file has the following entries:

header_block2d.txt
------------------
A header file containing the following information (line by line):

  - nt     the number of points in time in each entry
  - ny     the height of each 2d block
  - nx     the width of each 2d block
  - dtype  the used datatype (either "d" or "f")
  - name   the names of all 0d entries, each on its own line
  - 2      the indication that the 2d entries are starting
  - name   the names of all 2d entries, each on its own line

data.bin
--------
A binary file containing the following data, in the data type specified by the
header, and little-endian:

  - The nt time values
  - All 0d entries
  - All 2d entries, reshaped using numpy order='C'

""".strip()

# Encoding used for text portions of zip files
ENC = 'utf-8'


class DataBlock1d(object):
    """
    Container for time-series of 1d rectangular data arrays.

    Each ``DataBlock1d`` has a fixed width ``w``, and a 0d time series vector
    containing a sequence of ``n`` times.

    One-dimensional time series can be added to the block, provided the data
    also contains ``n`` instances of ``w`` data points. The data should be
    passed in as a numpy array with shape ``(n, w)``.

    Zero-dimensional time series can be added, provided they have length ``n``.

    A "one-dimensional time-series" is a series of equally sized 1d arrays
    (sequences), such that the first corresponds to a time ``t[0]``, the second
    to ``t[1]``, etc. Each array has shape ``(n, w)``.

    A "zero-dimensional time-series" is a series of single values where the
    first corresponds to a time ``t[0]``, the second to ``t[1]``, etc.

    Constructor info:

    ``w``
        Each 1d block should have dimension ``w`` by ``1``.
    ``time``
        A sequence of ``n`` times.
    ``copy``
        By default, a copy of the given time sequence will be stored. To
        prevent this copy, set ``copy=False``.

    """
    def __init__(self, w, time, copy=True):
        # Width
        w = int(w)
        if w < 1:
            raise ValueError('Minimum w is 1.')
        self._nx = w
        # Time
        time = np.array(time, copy=copy)
        if len(time.shape) != 1:
            raise ValueError('Time must be a sequence.')
        if np.any(np.diff(time) < 0):
            raise ValueError('Time must be non-decreasing.')
        self._time = time
        self._nt = len(time)
        # 0d variables
        self._0d = {}
        # 1d variables
        self._1d = {}

    def block2d(self):
        """
        Returns a :class:`myokit.DataBlock2d` based on this 1d data block.
        """
        b = DataBlock2d(self._nx, 1, self._time)
        for k, v in self._0d.items():
            b.set0d(k, v)
        shape = (self._nt, 1, self._nx)
        for k, v in self._1d.items():
            b.set2d(k, v.reshape(shape))
        return b

    def cv(self, name, threshold=-30, length=0.01, time_multiplier=1e-3,
            border=None):
        """
        Calculates conduction velocity (CV) in a cable.

        Accepts the following arguments:

        ``name``
            The name (as string) of the membrane potential variable. This
            should be a 1d variable in this datablock.
        ``threshold``
            The start of an action potential is determined as the first time
            the membrane potential crosses this threshold (default=-30mV) and
            has a positive direction.
        ``length``
            The length of a single cell in cm, in the direction of the cable.
            The default is ``length=0.01cm``.
        ``time_multiplier``
            A multiplier used to convert the used time units to seconds. Most
            simulations use milliseconds, so the default value is 1e-3.
        ``border``
            The number of cells to exclude from the analysis on each end of the
            cable to avoid boundary effects. If not given, 1/3 of the number of
            cells will be used, with a maximum of 50 cells on each side.

        Returns the approximate conduction velocity in cm/s. If no cv can be
        """
        # Check border
        if border is None:
            border = min(50, self._nx // 3)
        else:
            border = int(border)
            if border < 0:
                raise ValueError('The argument `border` cannot be negative.')
            elif border >= self._nx // 2:
                raise ValueError(
                    'The argument `border` must be less than half the number'
                    ' of cells.')

        # Get indices of selected cells
        ilo = border                # First indice
        ihi = self._nx - border     # Last indice + 1

        # Indices of cells with AP
        i1 = None
        i2 = None

        # Get Vm, reshaped to get each cell's time-series successively.
        v_series = self._1d[name].reshape(self._nt * self._nx, order='F')

        # Split Vm into a series per cell (returns views!)
        v_series = np.split(v_series, self._nx)

        # Find first activation time
        have_crossing = False
        t = []
        for i in range(ilo, ihi):
            v = v_series[i]
            # Get indice of first threshold crossing with positive flank
            # Don't include crossings at log indice 0
            itime = np.where((v[1:] > -30) & (v[1:] - v[:-1] > 0))[0]
            if len(itime) == 0 or itime[0] == 0:
                # No crossing found
                if have_crossing:
                    # CV calculation ends here
                    i2 = i - 1
                    break
            else:
                # Crossing found!
                if have_crossing:
                    i2 = i
                else:
                    i1 = i
                    have_crossing = True
                itime = 1 + itime[0]
                # Interpolate to get better estimate
                v0 = v[itime - 1]
                v1 = v[itime]
                t0 = self._time[itime - 1]
                t1 = self._time[itime]
                t.append(t0 + (threshold - v0) * (t1 - t0) / (v1 - v0))
        if not have_crossing:
            return 0

        # No propagation: all depolarisations at the same time
        if np.all(t == t[0]):
            return 0

        # Get times in seconds, lengths in cm
        t = np.array(t, copy=False) * time_multiplier
        x = np.arange(i1, 1 + i2, dtype=float) * length

        # Use linear least squares to find the conduction velocity
        A = np.vstack([t, np.ones(len(t))]).T
        try:
            # Newer numpy wants rcond=None
            cv = np.linalg.lstsq(A, x, rcond=None)[0][0]
        except TypeError:   # pragma: no cover
            cv = np.linalg.lstsq(A, x)[0][0]

        # Return
        return cv

    @staticmethod
    def from_DataLog(log):
        """
        Creates a DataBlock1d from a :class:`myokit.DataLog`.
        """
        log.validate()

        # Get time variable name
        time = log.time_key()
        if time is None:
            raise ValueError('No time variable set in data log.')

        # Get log info
        infos = log.variable_info()

        # Check time variable
        if time not in infos:
            # Already checked time variable exists, so if not found now must
            # be multi-dimensional
            raise ValueError('Time variable must be 0-dimensional.')

        # Check if everything is 0d or 1d, get size
        size = None
        for name, info in infos.items():
            d = info.dimension()
            if d not in (0, 1):
                raise ValueError(
                    'The given simulation log should only contain 0d or 1d'
                    ' variables. Found <' + str(name) + '> with d = '
                    + str(d) + '.')
            if d == 1:
                if size is None:
                    size = info.size()
                elif info.size() != size:
                    raise ValueError(
                        'The given simulation log contains 1d'
                        ' data sets of different sizes.')

        # Get dimensions
        nt = len(log[time])
        nx = size[0]

        # Create data block
        block = DataBlock1d(nx, log[time], copy=True)
        for name, info in infos.items():
            if info.dimension() == 0:
                # Add 0d time series
                if name == time:
                    continue
                block.set0d(name, log[name], copy=True)

            else:

                # Convert to 1d time series
                data = np.zeros(nt * nx)
                # Iterate over info.keys(), this has the correct order!
                for i, key in enumerate(info.keys()):
                    # Copy data into array (really copies)
                    data[i * nt:(i + 1) * nt] = log[key]
                # Reshape
                data = data.reshape((nt, nx), order='F')
                # If this is a view of existing data, make a copy!
                if data.base is not None:
                    data = np.array(data)
                block.set1d(name, data, copy=False)

        return block

    def get0d(self, name):
        """
        Returns the 0d time-series identified by ``name``. The data is returned
        directly, no copy is made.
        """
        return self._0d[name]

    def get1d(self, name):
        """
        Returns the 1d time-series identified by ``name``. The data is returned
        directly, no copy is made.

        The returned data is a 2d array of the shape given by :meth:`shape`.
        """
        return self._1d[name]

    def grid(self, name, transpose=True):
        """
        Returns a 2d grid representation suitable for plotting color maps or
        contours with ``matplotlib.pyplot`` methods such as ``pcolor`` and
        ``pcolormesh``.

        When used for example with
        ``pyplot.pcolormesh(*block.grid('membrane.V'))`` this will create a 2d
        plot where the horizontal axis shows time and the vertical axis shows
        the cell index.

        Arguments:

        ``name``
            The name identifying the 1d data-values to return.
        ``transpose``
            By default (``transpose=True``) the data is returned so that ``x``
            represents time and ``y`` represents space. To reverse this (and
            use the order used internally in the datablocks), set
            ``transpose=False``.

        The returned format is a tuple ``(x, y, z)`` where ``x``, ``y`` and
        ``z`` are all 2d numpy arrays.
        Here, ``x`` (time) and ``y`` (space) describe the x and y-coordinates
        of rectangles, with a color (data value) given by ``z``.

        In particular, each rectangle ``(x[i, j], y[i, j])``,
        ``(x[i + 1, j], y[i + 1, j])``, ``(x[i, j + 1], y[i, j + 1])``,
        ``(x[i + 1,j + 1], y[i + 1,j + 1])``, has a color given by ``z[i, j]``.

        As a result, for a block of width ``w`` (e.g., ``w`` cells) containing
        ``n`` logged time points, the method returns arrays ``x`` and ``y`` of
        shape ``(w + 1, n + 1)`` and an array ``z`` of shape ``(w, n)``.

        See :meth:`image_grid()` for a method where ``x``, ``y`` and ``z`` all
        have shape ``(w, n)``.
        """
        # Append point in time at pos [-1] + ([-1] - [-2])
        ts = np.append(self._time, 2 * self._time[-1] - self._time[-2])
        # Append one extra cell or node
        xs = np.arange(0, self._nx + 1)
        # Make grid
        return self._grid(ts, xs, self._1d[name], transpose)

    def image_grid(self, name, transpose=True):
        """
        Returns a 2d grid representation of the data.

        The returned format is a tuple ``(x, y, z)`` where ``x``, ``y`` and
        ``z`` are all 2d numpy arrays.
        Here, ``x`` and ``y`` describe the time and space-coordinates of the
        logged points respectively, and ``z`` describes the corresponding data
        value.
        For a block of width ``w`` (e.g., ``w`` cells) containing ``n`` logged
        time points, each returned array has the shape ``(w, n)``.

        Arguments:

        ``name``
            The name identifying the 1d data-values to return.
        ``transpose``
            By default, the data is transposed so that the ``x`` coordinates
            become time and the ``y`` coordinates become space. Use
            ``transpose=False`` to return untransposed results.

        """
        return self._grid(
            self._time, np.arange(0, self._nx), self._1d[name], transpose)

    def _grid(self, ts, xs, vs, transpose):
        """
        Make a grid for the given times, spatial coordinates and data values.
        """
        if transpose:
            x, y = np.meshgrid(ts, xs)
            z = np.reshape(vs, (self._nx * self._nt,), order='F')
            z = np.reshape(z, (self._nx, self._nt), order='C')
        else:
            x, y = np.meshgrid(xs, ts)
            z = np.reshape(vs, (self._nx * self._nt,), order='C')
            z = np.reshape(z, (self._nt, self._nx), order='C')
        # If z is a view, create a copy
        if z.base is not None:
            z = np.array(z, copy=True)
        return x, y, z

    def keys0d(self):
        """
        Returns an iterator over this block's 0d time series.
        """
        return iter(self._0d)

    def keys1d(self):
        """
        Returns an iterator over this block's 1d time series.
        """
        return iter(self._1d)

    def len0d(self):
        """
        Returns the number of 0d time series in this block.
        """
        return len(self._0d)

    def len1d(self):
        """
        Returns the number of 1d time series in this block.
        """
        return len(self._1d)

    @staticmethod
    def load(filename, progress=None, msg='Loading DataBlock1d'):
        """
        Loads a :class:`DataBlock1d` from the specified file.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        filename = os.path.expanduser(filename)

        # Load compression modules
        import zipfile
        try:
            import zlib
            del(zlib)
        except ImportError:
            raise Exception(
                'This method requires the ``zlib`` module to be installed.')

        # Get size of single and double types on this machine
        try:
            dsize = {
                'd': len(array.array('d', [1]).tobytes()),
                'f': len(array.array('f', [1]).tobytes()),
            }
        except (AttributeError, TypeError):  # pragma: no python 3 cover
            # List dtype as str for Python 2.7.10 (see #225)
            dsize = {
                b'd': len(array.array(b'd', [1]).tostring()),
                b'f': len(array.array(b'f', [1]).tostring()),
            }

        # Read data from file
        try:
            f = None
            f = zipfile.ZipFile(filename, 'r')
            info = f.infolist()
            if len(info) < 3:
                raise myokit.DataBlockReadError(
                    'Invalid DataBlock1d file format: Not enough files in'
                    ' zip.')

            # Get ZipInfo objects
            names = [x.filename for x in info]
            try:
                head = names.index('header_block1d.txt')
            except ValueError:
                raise myokit.DataBlockReadError(
                    'Invalid DataBlock1d file format: Header not found.')
            try:
                body = names.index('data.bin')
            except ValueError:
                raise myokit.DataBlockReadError(
                    'Invalid DataBlock1d file format: Data not found.')

            # Read head and body into memory (let's assume it fits...)
            head = f.read(info[head]).decode(ENC)
            body = f.read(info[body])

        except zipfile.BadZipfile:
            raise myokit.DataBlockReadError(
                'Unable to read DataBlock1d: Bad zip file.')

        except zipfile.LargeZipFile:    # pragma: no cover
            raise myokit.DataBlockReadError(
                'Unable to read DataBlock1d: Zip file requires zip64 support'
                ' and this has not been enabled on this system.')

        finally:
            if f:
                f.close()

        # Parse head
        head = head.splitlines()
        try:
            if progress:
                progress.enter(msg)

                # Avoid divide by zero
                fraction = float(len(head) - 3)
                if fraction > 0:
                    fraction = 1.0 / fraction
                iprogress = 0
                progress.update(iprogress * fraction)
            head = iter(head)
            nt = int(next(head))
            nx = int(next(head))
            dtype = next(head)[1:-1]
            # Convert dtype to str for Python 2.7.10 (see #225)
            dtype = str(dtype)
            if dtype not in dsize:
                raise myokit.DataBlockReadError(
                    'Unable to read DataBlock1d: Unrecognized data type "'
                    + dtype + '".')
            names_0d = []
            names_1d = []
            name = next(head)
            while name != '1':
                names_0d.append(name[1:-1])
                name = next(head)
            for name in head:
                names_1d.append(name[1:-1])
            del(head)

            # Parse body
            start, end = 0, 0
            n0 = dsize[dtype] * nt
            n1 = n0 * nx
            nb = len(body)

            # Read time
            end += n0
            if end > nb:
                raise myokit.DataBlockReadError(
                    'Unable to read DataBlock1d: Header indicates larger data'
                    ' than found in the body.')
            data = array.array(dtype)
            try:
                data.frombytes(body[start:end])
            except AttributeError:  # pragma: no python 3 cover
                data.fromstring(body[start:end])
            if sys.byteorder == 'big':  # pragma: no cover
                data.byteswap()
            data = np.array(data)
            if progress:
                iprogress += 1
                if not progress.update(iprogress * fraction):
                    return

            # Create data block
            block = DataBlock1d(nx, data, copy=False)

            # Read 0d data
            for name in names_0d:
                start = end
                end += n0
                if end > nb:
                    raise myokit.DataBlockReadError(
                        'Unable to read DataBlock1d: Header indicates larger'
                        ' data than found in the body.')
                data = array.array(dtype)
                try:
                    data.frombytes(body[start:end])
                except AttributeError:  # pragma: no python 3 cover
                    data.fromstring(body[start:end])
                if sys.byteorder == 'big':  # pragma: no cover
                    data.byteswap()
                data = np.array(data)
                block.set0d(name, data, copy=False)
                if progress:
                    iprogress += 1
                    if not progress.update(iprogress * fraction):
                        return

            # Read 1d data
            for name in names_1d:
                start = end
                end += n1
                if end > nb:
                    raise myokit.DataBlockReadError(
                        'Unable to read DataBlock1d: Header indicates larger'
                        ' data than found in the body.')
                data = array.array(dtype)
                try:
                    data.frombytes(body[start:end])
                except AttributeError:  # pragma: no python 3 cover
                    data.fromstring(body[start:end])
                if sys.byteorder == 'big':  # pragma: no cover
                    data.byteswap()
                data = np.array(data).reshape(nt, nx, order='C')
                block.set1d(name, data, copy=False)
                if progress:
                    iprogress += 1
                    if not progress.update(iprogress * fraction):
                        return
            return block
        finally:
            if progress:
                progress.exit()

    def save(self, filename):
        """
        Writes this ``DataBlock1d`` to a binary file.

        The resulting file will be a zip file with the following entries:

        ``header_block1d.txt``: A header file containing the following
        information (line by line):

        - ``nt`` the number of points in time in each entry
        - ``nx`` the length of each 1d block
        - ``"dtype"`` the used datatype (either "d" or "f")
        - ``"name"`` the names of all 0d entries, each on its own line
        - ``1`` the indication that the 1d entries are starting
        - ``"name"`` the names of all 1d entries, each on its own line

        ``data.bin``: A binary file containing the following data, in the data
        type specified by the header, and little-endian:

        - The ``nt`` time values
        - All 0d entries
        - All 1d entries, reshaped using numpy order='C'

        """
        # Check filename
        filename = os.path.expanduser(filename)
        # Load compression modules
        import zipfile
        try:
            import zlib
            del(zlib)
        except ImportError:
            raise Exception(
                'This method requires the ``zlib`` module to be installed.')

        # Data type
        # Create dtype as str for Python 2.7.10 (see #225)
        dtype = str('d')     # Only supporting doubles right now

        # Create header
        head_str = []
        head_str.append(str(self._nt))
        head_str.append(str(self._nx))
        head_str.append('"' + dtype + '"')
        for name in self._0d:
            head_str.append('"' + name + '"')
        head_str.append(str(1))
        for name in self._1d:
            head_str.append('"' + name + '"')
        head_str = '\n'.join(head_str)

        # Create body
        n = self._nt * self._nx
        body_str = []
        body_str.append(array.array(dtype, self._time))
        for name, data in self._0d.items():
            body_str.append(array.array(dtype, data))
        for name, data in self._1d.items():
            body_str.append(array.array(dtype, data.reshape(n, order='C')))
        if sys.byteorder == 'big':  # pragma: no cover
            for ar in body_str:
                ar.byteswap()
        try:
            body_str = b''.join([ar.tobytes() for ar in body_str])
        except AttributeError:  # pragma: no python 3 cover
            body_str = b''.join([ar.tostring() for ar in body_str])

        # Write
        head = zipfile.ZipInfo('header_block1d.txt')
        head.compress_type = zipfile.ZIP_DEFLATED
        body = zipfile.ZipInfo('data.bin')
        body.compress_type = zipfile.ZIP_DEFLATED
        read = zipfile.ZipInfo('readme.txt')
        read.compress_type = zipfile.ZIP_DEFLATED
        with zipfile.ZipFile(filename, 'w') as f:
            f.writestr(head, head_str.encode(ENC))
            f.writestr(body, body_str)
            f.writestr(read, README_SAVE_1D.encode(ENC))

    def set0d(self, name, data, copy=True):
        """
        Adds or updates a zero-dimensional time series ``data`` for the
        variable named by the string ``name``.

        The ``data`` must be specified as a sequence of length ``n``, where
        ``n`` is the first value returned by :meth:`DataBlock1d.shape()`.

        By default, a copy of the given data will be stored. To prevent this
        and store a reference instead, set ``copy=False``.
        """
        name = str(name)
        if not name:
            raise ValueError('Name cannot be empty.')
        data = np.array(data, copy=copy)
        if data.shape != (self._nt,):
            raise ValueError(
                'Data must be sequence of length ' + str(self._nt) + '.')
        self._0d[name] = data

    def set1d(self, name, data, copy=True):
        """
        Adds or updates a one-dimensional time series ``data`` for the variable
        named by the string ``name``.

        The ``data`` must be specified as a numpy array with shape ``(n, w)``,
        where ``(n, w)`` is the value returned by :meth:`DataBlock1d.shape()`.

        By default, a copy of the given data will be stored. To prevent this
        and store a reference instead, set ``copy=False``.
        """
        name = str(name)
        if not name:
            raise ValueError('Name cannot be empty.')
        data = np.array(data, copy=copy)
        shape = (self._nt, self._nx)
        if data.shape != shape:
            raise ValueError('Data must have shape ' + str(shape) + '.')
        self._1d[name] = data

    def shape(self):
        """
        Returns the required shape for 1d data passed to this data block. Zero
        dimensional series passed in must have length ``shape()[0]``.
        """
        return (self._nt, self._nx)

    def time(self):
        """
        Returns the time data for this datablock. The data is returned
        directly, no copy is made.
        """
        return self._time

    def trace(self, variable, x):
        """
        Returns a 0d time series of the value ``variable``, corresponding to
        the cell at position ``x``. The data is returned directly, no copy is
        made.
        """
        return self._1d[variable][:, x]


class DataBlock2d(object):
    """
    Container for time-series of 2d rectangular data arrays.

    Each ``DataBlock2d`` has a fixed width ``w`` and height ``h``, and a 0d
    time series vector containing a sequence of ``n`` times.

    Two-dimensional time series can be added to the block, provided the data
    also contains ``n`` instances of ``w`` by ``h`` data points. The
    data should be passed in as a numpy array with shape ``(n, h, w)``.

    Zero-dimensional time series can be added, provided they have length ``n``.

    A "two-dimensional time-series" is a series of equally sized 2d arrays
    (sequences), such that the first corresponds to a time ``t[0]``, the second
    to ``t[1]``, etc.

    A "zero-dimensional time-series" is a series of single values where the
    first corresponds to a time ``t[0]``, the second to ``t[1]``, etc.

    Constructor info:

    ``w``
        The width of a 2d block. Each block should have shape (n, h, w)
    ``h``
        The height of a 2d block. Each block should have shape (n, h, w)
    ``time``
        A sequence of ``n`` times.
    ``copy``
        By default, a copy of the given time sequence will be stored. To
        prevent this copy, set ``copy=False``.

    """
    def __init__(self, w, h, time, copy=True):
        # Width and height
        w, h = int(w), int(h)
        if w < 1:
            raise ValueError('Minimum width is 1.')
        if h < 1:
            raise ValueError('Minimum height is 1.')
        self._ny = h
        self._nx = w
        # Time
        time = np.array(time, copy=copy)
        if len(time.shape) != 1:
            raise ValueError('Time must be a sequence.')
        if not np.all(np.diff(time) >= 0):
            raise ValueError('Time must be non-decreasing.')
        self._time = time
        self._nt = len(time)
        # 0d variables
        self._0d = {}
        # 2d variables
        self._2d = {}

    def colors(self, name, colormap='traditional', lower=None, upper=None):
        """
        Converts the 2d series indicated by ``name`` into a list of ``W*H*RGB``
        arrays, with each entry represented as an 8 bit unsigned integer.
        """
        data = self._2d[name]
        # Get color map
        color_map = ColorMap.get(colormap)
        # Get lower and upper bounds for colormap scaling
        lower = np.min(data) if lower is None else float(lower)
        upper = np.max(data) if upper is None else float(upper)
        # Create images
        frames = []
        for frame in data:
            # Convert 2d array into row-strided array
            frame = frame.reshape(self._ny * self._nx, order='C')
            # Apply colormap
            frame = color_map(
                frame, lower=lower, upper=upper, alpha=False, rgb=True)
            # Reshape to nx * ny * 3 color array
            frame = frame.reshape((self._ny, self._nx, 3))
            # Append to list
            frames.append(frame)
        return frames

    @staticmethod
    def combine(block1, block2, map2d, map0d=None, pos1=None, pos2=None):
        """
        Combines two blocks, containing information about different areas, into
        a single :class:`DataBlock2d`.

        Both blocks must contain data from the same points in time.

        A mapping from old to new variables must be passed in as a dictionary
        ``map2d``. The blocks can have different sizes but must have the same
        time vector. If any empty space is created it is padded with a value
        taken from one of the data blocks or a padding value specified as part
        of ``map2d``.

        Positions for the datablocks can be specified as ``pos1`` and ``pos2``,
        the new datablock will have indices ranging from ``(0, 0)`` to
        ``(max(pos1[0] + w1, pos2[0] + w2), max(pos1[0] + w1, pos2[0] + w2))``,
        where ``w1`` and ``w2`` are the widths of ``block1`` and ``block2``
        respectively. Negative indices are not supported and the blocks are not
        allowed to overlap.

        Arguments:

        ``block1``
            The first DataBlock2d
        ``block2``
            The second DataBlock2d. This must have the same time vector as the
            first.
        ``map2d``
            A dictionary object showing how to map 2d variables from both
            blocks into the newly created datablock. The format must be:
            ``new_name : (old_name_1, old_name_2, padding_value)``. Here,
            ``new_name`` is the name of the new 2d variable, ``old_name_1`` is
            the name of a 2d variable in ``block1``, ``old_name_2`` is the name
            of a 2d variable in ``block2`` and ``padding_value`` is an optional
            value indicating the value to use for undefined spaces in the new
            block.
        ``map0d=None``,
            A dictionary object showing how to map 0d variables from both
            blocks into the newly created datablock. Each entry must take the
            format: ``new_name : (old_name_1, None)`` or
            ``new_name : (None, old_name_2)``.
        ``pos1=None``
            Optional value indicating the position ``(x, y)`` of the first
            datablock. By default ``(0, 0)`` is used.
        ``pos2=None``
            Optional value indicating the position ``(x, y)`` of the first
            datablock. By default ``(w1, 0)`` is used, where ``w1`` is the
            width of ``block1``.

        """
        # Check time vector
        time = block1.time()
        if not np.allclose(time, block2.time()):
            raise ValueError(
                'Both datablocks must contain data from the same points in'
                ' time.')

        # Check indices
        nt, h1, w1 = block1.shape()
        nt, h2, w2 = block2.shape()

        if pos1:
            x1, y1 = [int(i) for i in pos1]
            if x1 < 0 or y1 < 0:
                raise ValueError(
                    'Negative indices not supported: pos1=('
                    + str(x1) + ', ' + str(y1) + ').')
        else:
            x1, y1 = 0, 0

        if pos2:
            x2, y2 = [int(i) for i in pos2]
            if x2 < 0 or y2 < 0:
                raise ValueError(
                    'Negative indices not supported: pos2=('
                    + str(x2) + ', ' + str(y2) + ').')
        else:
            x2, y2 = x1 + w1, 0

        # Check for overlap
        if not (x1 >= x2 + w2 or x2 >= x1 + w1
                or y1 >= y2 + h2 or y2 >= y1 + h2):
            raise ValueError('The two data blocks indices cannot overlap.')

        # Create new datablock
        nx = max(x1 + w1, x2 + w2)
        ny = max(y1 + h1, y2 + h2)
        block = DataBlock2d(nx, ny, time, copy=True)

        # Enter 0d data
        if map0d:
            for name, old in map0d.items():
                if old[0] is None:
                    b = block2
                    n = old[1]
                elif old[1] is None:
                    b = block1
                    n = old[0]
                else:
                    raise ValueError(
                        'The dictionary map0d must map the names of new 0d'
                        ' entries to a tuple (a, b) where either a or b must'
                        ' be None.')
                block.set0d(name, b.get0d(n))

        # Enter 2d data
        for name, source in map2d.items():
            # Get data sources
            name1, name2 = source[0], source[1]
            source1 = block1.get2d(name1)
            source2 = block2.get2d(name2)

            # Get padding value
            try:
                pad = float(source[2])
            except IndexError:
                # Get lowest value
                pad = min(np.min(source1), np.min(source2))

            # Create new data field
            field = pad * np.ones((nt, ny, nx))
            field[:, y1:y1 + h1, x1:x1 + w1] = source1
            field[:, y2:y2 + h2, x2:x2 + w2] = source2
            block.set2d(name, field)

        # Return new block
        return block

    def dominant_eigenvalues(self, name):
        """
        Takes the 2d data specified by ``name`` and computes the dominant
        eigenvalue for each point in time. This only works for datablocks with
        a square 2d grid.

        The "dominant eigenvalue" is defined as the eigenvalue with the largest
        magnitude (``sqrt(a + bi)``).

        The returned data is a 1d numpy array.
        """
        if self._nx != self._ny:
            raise Exception(
                'Eigenvalues can only be determined for square data blocks.')
        data = self._2d[name]
        dominants = []
        for t in range(self._nt):
            e = np.linalg.eigvals(data[t])
            dominants.append(e[np.argmax(np.absolute(e))])
        return np.array(dominants)

    def eigenvalues(self, name):
        """
        Takes the 2d data specified as ``name`` and computes the eigenvalues of
        its data matrix at every point in time. This only works for datablocks
        with a square 2d grid.

        The returned data is a 2d numpy array where the first axis is time and
        the second axis is the index of each eigenvalue.
        """
        if self._nx != self._ny:
            raise Exception(
                'Eigenvalues can only be determined for square data blocks.')
        data = self._2d[name]
        eigenvalues = []
        for t in range(self._nt):
            eigenvalues.append(np.linalg.eigvals(data[t]))
        return np.array(eigenvalues)

    @staticmethod
    def from_DataLog(log):
        """
        Creates a DataBlock2d from a :class:`myokit.DataLog`.
        """
        log.validate()

        # Get time variable name
        time = log.time_key()
        if time is None:
            raise ValueError('No time variable set in data log.')

        # Get log info
        infos = log.variable_info()

        # Check time variable
        if time not in infos:
            # Already checked time variable exists, so if not found now must
            # be multi-dimensional
            raise ValueError('Time variable must be 0-dimensional.')

        # Check if everything is 0d or 2d, get size
        size = None
        for name, info in infos.items():
            d = info.dimension()
            if d not in (0, 2):
                raise ValueError(
                    'The given simulation log should only contain 0d or 2d'
                    ' variables. Found <' + str(name) + '> with d = '
                    + str(d) + '.')
            if d == 2:
                if size is None:
                    size = info.size()
                elif info.size() != size:
                    raise ValueError(
                        'The given simulation log contains 2d data sets of'
                        ' different sizes.')
        # Get dimensions
        nt = len(log[time])
        nx, ny = size
        # Create data block
        block = DataBlock2d(nx, ny, log[time], copy=True)
        for name, info in infos.items():
            if info.dimension() == 0:
                # Add 0d time series
                if name == time:
                    continue
                block.set0d(name, log[name], copy=True)
            else:
                # Convert to 2d time series
                data = np.zeros(nt * ny * nx)
                # Iterate over info.keys()
                for i, key in enumerate(info.keys()):
                    # Copy data into array (really copies)
                    data[i * nt:(i + 1) * nt] = log[key]
                # Reshape
                data = data.reshape((nt, ny, nx), order='F')
                # If this is a view of existing data, make a copy!
                if data.base is not None:
                    data = np.array(data)
                block.set2d(name, data, copy=False)
        return block

    def get0d(self, name):
        """
        Returns the 0d time-series identified by ``name``. The data is returned
        directly, no copy is made.
        """
        return self._0d[name]

    def get2d(self, name):
        """
        Returns the 2d time-series identified by ``name``. The data is returned
        directly, no copy is made.
        """
        return self._2d[name]

    def images(self, name, colormap='traditional', lower=None, upper=None):
        """
        Converts the 2d series indicated by ``name`` into a list of 1d arrays
        in a row-strided image format ``ARGB32``.
        """
        data = self._2d[name]
        # Get color map
        color_map = ColorMap.get(colormap)
        # Get lower and upper bounds for colormap scaling
        lower = np.min(data) if lower is None else float(lower)
        upper = np.max(data) if upper is None else float(upper)
        # Create images
        frames = []
        for frame in data:
            # Convert 2d array into row-strided array
            frame = frame.reshape(self._ny * self._nx, order='C')
            frames.append(color_map(frame, lower=lower, upper=upper))
        return frames

    def is_square(self):
        """
        Returns True if this data block's grid is square.
        """
        return self._nx == self._ny

    def items0d(self):
        """
        Returns an iterator over ``(name, value)`` pairs for the 0d series
        stored in this block. The given values are references! No copy of the
        data is made.
        """
        return self._0d.items()

    def items2d(self):
        """
        Returns an iterator over ``(name, value)`` pairs for the 2d series
        stored in this block. The given values are references! No copy of the
        data is made.
        """
        return self._2d.items()

    def keys0d(self):
        """
        Returns an iterator over this block's 0d time series.
        """
        return iter(self._0d)

    def keys2d(self):
        """
        Returns an iterator over this block's 2d time series.
        """
        return iter(self._2d)

    def largest_eigenvalues(self, name):
        """
        Takes the 2d data specified by ``name`` and computes the largest
        eigenvalue for each point in time. This only works for datablocks with
        a square 2d grid.

        The "largest eigenvalue" is defined as the eigenvalue with the most
        positive real part. Note that the returned values may be complex.

        The returned data is a 1d numpy array.
        """
        if self._nx != self._ny:
            raise Exception(
                'Eigenvalues can only be determined for square data blocks.')
        data = self._2d[name]
        largest = []
        for t in range(self._nt):
            e = np.linalg.eigvals(data[t])
            largest.append(e[np.argmax(np.real(e))])
        return np.array(largest)

    def len0d(self):
        """
        Returns the number of 0d time series in this block.
        """
        return len(self._0d)

    def len2d(self):
        """
        Returns the number of 2d time series in this block.
        """
        return len(self._2d)

    @staticmethod
    def load(filename, progress=None, msg='Loading DataBlock2d'):
        """
        Loads a :class:`DataBlock2d` from the specified file.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.

        If the given file contains a :class:`DataBlock1d` this is read and
        converted to a 2d block without warning.
        """
        filename = os.path.expanduser(filename)

        # Load compression modules
        import zipfile
        try:
            import zlib
            del(zlib)
        except ImportError:
            raise Exception(
                'This method requires the ``zlib`` module to be installed.')

        # Get size of single and double types on this machine
        try:
            dsize = {
                'd': len(array.array('d', [1]).tobytes()),
                'f': len(array.array('f', [1]).tobytes()),
            }
        except (AttributeError, TypeError):  # pragma: no python 3 cover
            # List dtype as str for Python 2.7.10 (see #225)
            dsize = {
                b'd': len(array.array(b'd', [1]).tostring()),
                b'f': len(array.array(b'f', [1]).tostring()),
            }

        # Read data from file
        try:
            f = None
            f = zipfile.ZipFile(filename, 'r')
            info = f.infolist()
            if len(info) < 3:
                raise myokit.DataBlockReadError(
                    'Invalid DataBlock2d file format: Not enough files in'
                    ' zip.')

            # Get ZipInfo objects
            names = [x.filename for x in info]
            try:
                head = names.index('header_block2d.txt')
            except ValueError:
                # Attempt reading as DataBlock1d
                try:
                    names.index('header_block1d.txt')
                except ValueError:
                    raise myokit.DataBlockReadError(
                        'Invalid DataBlock2d file format: Header not found.')

                # It's a DataBlock1d, attempt reading as such
                f.close()
                block1d = DataBlock1d.load(filename, progress, msg)
                return None if block1d is None else block1d.block2d()

            try:
                body = names.index('data.bin')
            except ValueError:
                raise myokit.DataBlockReadError(
                    'Invalid DataBlock2d file format: Data not found.')

            # Read head and body into memory (let's assume it fits...)
            head = f.read(info[head]).decode(ENC)
            body = f.read(info[body])

        except zipfile.BadZipfile:
            raise myokit.DataBlockReadError(
                'Unable to read DataBlock2d: Bad zip file.')

        except zipfile.LargeZipFile:    # pragma: no cover
            raise myokit.DataBlockReadError(
                'Unable to read DataBlock2d: zip file requires zip64 support'
                ' and this has not been enabled on this system.')

        finally:
            if f:
                f.close()

        # Parse head
        head = head.splitlines()
        try:
            if progress:
                progress.enter(msg)
                # Avoid divide by zero
                fraction = float(len(head) - 4)
                if fraction > 0:
                    fraction = 1.0 / fraction
                iprogress = 0
                progress.update(iprogress * fraction)
            head = iter(head)
            nt = int(next(head))
            ny = int(next(head))
            nx = int(next(head))

            # Get dtype
            # Convert dtype to str for Python 2.7.10 (see #225)
            dtype = str(next(head))[1:-1]
            if dtype not in dsize:
                raise myokit.DataBlockReadError(
                    'Unable to read DataBlock2d: Unrecognized data type "'
                    + str(dtype) + '".')

            names_0d = []
            names_2d = []
            name = next(head)
            while name != '2':
                names_0d.append(name[1:-1])
                name = next(head)
            for name in head:
                names_2d.append(name[1:-1])
            del(head)

            # Parse body
            start, end = 0, 0
            n0 = dsize[dtype] * nt
            n2 = n0 * ny * nx
            nb = len(body)

            # Read time
            end += n0
            if end > nb:
                raise myokit.DataBlockReadError(
                    'Unable to read DataBlock2d: Header indicates larger data'
                    ' than found in the body.')

            data = array.array(dtype)
            try:
                data.frombytes(body[start:end])
            except AttributeError:  # pragma: no python 3 cover
                data.fromstring(body[start:end])
            if sys.byteorder == 'big':  # pragma: no cover
                data.byteswap()
            data = np.array(data)
            if progress:
                iprogress += 1
                if not progress.update(iprogress * fraction):
                    return

            # Create data block
            block = DataBlock2d(nx, ny, data, copy=False)

            # Read 0d data
            for name in names_0d:
                start = end
                end += n0
                if end > nb:
                    raise myokit.DataBlockReadError(
                        'Unable to read DataBlock2d: Header indicates larger'
                        ' data than found in the body.')
                data = array.array(dtype)
                try:
                    data.frombytes(body[start:end])
                except AttributeError:  # pragma: no python 3 cover
                    data.fromstring(body[start:end])
                if sys.byteorder == 'big':  # pragma: no cover
                    data.byteswap()
                data = np.array(data)
                block.set0d(name, data, copy=False)
                if progress:
                    iprogress += 1
                    if not progress.update(iprogress * fraction):
                        return

            # Read 2d data
            for name in names_2d:
                start = end
                end += n2
                if end > nb:
                    raise myokit.DataBlockReadError(
                        'Unable to read DataBlock2d: Header indicates larger'
                        ' data than found in the body.')
                data = array.array(dtype)
                try:
                    data.frombytes(body[start:end])
                except AttributeError:  # pragma: no python 3 cover
                    data.fromstring(body[start:end])
                if sys.byteorder == 'big':  # pragma: no cover
                    data.byteswap()
                data = np.array(data).reshape(nt, ny, nx, order='C')
                block.set2d(name, data, copy=False)
                if progress:
                    iprogress += 1
                    if not progress.update(iprogress * fraction):
                        return
            return block

        finally:

            if progress:
                progress.exit()

    def save(self, filename):
        """
        Writes this ``DataBlock2d`` to a binary file.

        The resulting file will be a zip file with the following entries:

        ``header_block2d.txt``: A header file containing the following
        information (line by line):

        - ``nt`` the number of points in time in each entry
        - ``ny`` the height of each 2d block
        - ``nx`` the width of each 2d block
        - ``"dtype"`` the used datatype (either "d" or "f")
        - ``"name"`` the names of all 0d entries, each on its own line
        - ``2`` the indication that the 2d entries are starting
        - ``"name"`` the names of all 2d entries, each on its own line

        ``data.bin``: A binary file containing the following data, in the data
        type specified by the header, and little-endian:

        - The ``nt`` time values
        - All 0d entries
        - All 2d entries, reshaped using numpy order='C'

        """
        # Check filename
        filename = os.path.expanduser(filename)

        # Load compression modules
        import zipfile
        try:
            # Check zlib is available
            import zlib
            del(zlib)
        except ImportError:
            raise Exception(
                'This method requires the ``zlib`` module to be installed.')

        # Data type
        # Create dtype as str for Python 2.7.10 (see #225)
        dtype = str('d')  # Only supporting doubles right now

        # Create header
        head_str = []
        head_str.append(str(self._nt))
        head_str.append(str(self._ny))
        head_str.append(str(self._nx))
        head_str.append('"' + dtype + '"')
        for name in self._0d:
            head_str.append('"' + name + '"')
        head_str.append(str(2))
        for name in self._2d:
            head_str.append('"' + name + '"')
        head_str = '\n'.join(head_str)

        # Create body
        n = self._nt * self._ny * self._nx
        body_str = []
        body_str.append(array.array(dtype, self._time))
        for name, data in self._0d.items():
            body_str.append(array.array(dtype, data))
        for name, data in self._2d.items():
            body_str.append(array.array(dtype, data.reshape(n, order='C')))
        if sys.byteorder == 'big':  # pragma: no cover
            for ar in body_str:
                ar.byteswap()
        try:
            body_str = b''.join([ar.tobytes() for ar in body_str])
        except AttributeError:  # pragma: no python 3 cover
            body_str = b''.join([ar.tostring() for ar in body_str])

        # Write
        head = zipfile.ZipInfo('header_block2d.txt')
        head.compress_type = zipfile.ZIP_DEFLATED
        body = zipfile.ZipInfo('data.bin')
        body.compress_type = zipfile.ZIP_DEFLATED
        read = zipfile.ZipInfo('readme.txt')
        read.compress_type = zipfile.ZIP_DEFLATED
        with zipfile.ZipFile(filename, 'w') as f:
            f.writestr(head, head_str.encode(ENC))
            f.writestr(body, body_str)
            f.writestr(read, README_SAVE_2D.encode(ENC))

    def save_frame_csv(
            self, filename, name, frame, xname='x', yname='y', zname='value'):
        """
        Stores a single 2d variable's data at a single point in time to disk,
        using a csv format where each point in the frame is stored on a
        separate line as a tuple ``x, y, value``.
        """
        # Check filename
        filename = os.path.expanduser(filename)
        # Save
        delimx = ','
        delimy = '\n'
        data = self._2d[name]
        data = data[frame]
        text = [delimx.join('"' + str(x) + '"' for x in [xname, yname, zname])]
        for y, row in enumerate(data):
            for x, z in enumerate(row):
                text.append(delimx.join([str(x), str(y), myokit.strfloat(z)]))
        text = delimy.join(text)
        with open(filename, 'w') as f:
            f.write(text)

    def save_frame_grid(self, filename, name, frame, delimx=' ', delimy='\n'):
        """
        Stores a single 2d variable's data at a single point in time to disk,
        using a simple 2d format where each row of the resulting data file
        represents a row of the frame.

        Data from 2d variable ``name`` at frame ``frame`` will be stored in
        ``filename`` row by row. Each column is separated by ``delimx`` (by
        default a space) and rows are separated by ``delimy`` (by default this
        will be a newline character).
        """
        # Check filename
        filename = os.path.expanduser(filename)
        # Save
        data = self._2d[name]
        data = data[frame]
        text = []
        for row in data:
            text.append(delimx.join([myokit.strfloat(x) for x in row]))
        text = delimy.join(text)
        with open(filename, 'w') as f:
            f.write(text)

    def set0d(self, name, data, copy=True):
        """
        Adds or updates a zero-dimensional time series ``data`` for the
        variable named by the string ``name``.

        The ``data`` must be specified as a sequence of length ``n``, where
        ``n`` is the first value returned by :meth:`DataBlock2d.shape()`.

        By default, a copy of the given data will be stored. To prevent this
        and store a reference instead, set ``copy=False``.
        """
        name = str(name)
        if not name:
            raise ValueError('Name cannot be empty.')
        data = np.array(data, copy=copy)
        if data.shape != (self._nt,):
            raise ValueError(
                'Data must be sequence of length ' + str(self._nt) + '.')
        self._0d[name] = data

    def set2d(self, name, data, copy=True):
        """
        Adds or updates a two-dimensional time series ``data`` for the variable
        named by the string ``name``.

        The ``data`` must be specified as a numpy array with shape ``(n, w)``,
        where ``(n, w)`` is the value returned by :meth:`DataBlock2d.shape()`.

        By default, a copy of the given data will be stored. To prevent this
        and store a reference instead, set ``copy=False``.
        """
        name = str(name)
        if not name:
            raise ValueError('Name cannot be empty.')
        data = np.array(data, copy=copy)
        shape = (self._nt, self._ny, self._nx)
        if data.shape != shape:
            raise ValueError('Data must have shape ' + str(shape) + '.')
        self._2d[name] = data

    def shape(self):
        """
        Returns the required shape for 2d data passed to this data block. Zero
        dimensional series passed in must have length ``shape()[0]``.
        """
        return (self._nt, self._ny, self._nx)

    def time(self):
        """
        Returns the time data for this datablock. The data is returned
        directly, no copy is made.
        """
        return self._time

    def trace(self, variable, x, y):
        """
        Returns a 0d time series of the value ``variable``, corresponding to
        the cell at position ``x``, ``y``. The data is returned directly, no
        copy is made.
        """
        return self._2d[variable][:, y, x]


class ColorMap(object):
    """
    *Abstract class*

    Applies colormap transformations to floating point data and returns RGB
    data.

    :class:`ColorMaps <ColorMap>` are callable objects and take the following
    arguments:

    ``floats``
        A 1-dimensional numpy array of floating point numbers.
    ``lower=None``
        A lower bound for the floats in the input. The ``lower`` and ``upper``
        values are used to normalize the input before applying the colormap. If
        this bound is omitted the lowest value in the input data is used.
    ``upper=None``
        An upper bound for the floats in the input. The ``lower`` and ``upper``
        values are used to normalize the input before applying the colormap. If
        this bound is omitted the highest value in the input data is used.
    ``alpha=True``
        Set to ``False`` to omit an alpha channel from the output.
    ``rgb=None``
        Set to ``True`` to return bytes in the order ``0xARGB``, to ``False``
        to return the order ``0xBGRA`` or to ``None`` to let the system's
        endianness determine the correct order. In the last case, big-endian
        systems will return ``0xARGB`` while little-endian systems use the
        order ``0xBGRA``.

    A 1-dimensional array of ``n`` floating point numbers will be converted to
    a 1-dimensional array of ``4n`` ``uints``, or ``3n`` if the alpha channel
    is disabled. The array will be ordered sequentially: the first four (or
    three) bytes describe the first float, the next four (or three) describe
    the second float and so on.
    """
    _colormaps = {}

    def __call__(self, floats, lower=None, upper=None, alpha=True, rgb=None):
        raise NotImplementedError

    @staticmethod
    def exists(name):
        """
        Returns True if the given name corresponds to a colormap.
        """
        return name in ColorMap._colormaps

    @staticmethod
    def get(name):
        """
        Returns the colormap method indicated by the given name.
        """
        try:
            return ColorMap._colormaps[name]()
        except KeyError:
            raise KeyError('Non-existent ColorMap "' + str(name) + '".')

    @staticmethod
    def hsv_to_rgb(h, s, v):
        """
        Converts hsv values in the range [0,1] to rgb values in the range
        [0,255]. Adapted from Matplotlib.
        """
        r, g, b = np.empty_like(h), np.empty_like(h), np.empty_like(h)
        i = (h * 6).astype(np.int) % 6
        f = (h * 6) - i
        p = v * (1 - s)
        q = v * (1 - s * f)
        t = v * (1 - s * (1 - f))
        idx = (i == 0)
        r[idx], g[idx], b[idx] = v[idx], t[idx], p[idx]
        idx = (i == 1)
        r[idx], g[idx], b[idx] = q[idx], v[idx], p[idx]
        idx = (i == 2)
        r[idx], g[idx], b[idx] = p[idx], v[idx], t[idx]
        idx = (i == 3)
        r[idx], g[idx], b[idx] = p[idx], q[idx], v[idx]
        idx = (i == 4)
        r[idx], g[idx], b[idx] = t[idx], p[idx], v[idx]
        idx = (i == 5)
        r[idx], g[idx], b[idx] = v[idx], p[idx], q[idx]
        out = (
            np.array(r * 255, dtype=np.uint8, copy=False),
            np.array(g * 255, dtype=np.uint8, copy=False),
            np.array(b * 255, dtype=np.uint8, copy=False),
        )
        return out

    @staticmethod
    def image(name, x, y):
        """
        Returns image data (such as returned by :meth:`DataBlock2d.images()`)
        representing the colormap specified by ``name``. The image dimensions
        can be set using ``x`` and ``y``.
        """
        data = np.linspace(1, 0, y)
        data = np.tile(data, (x, 1)).transpose()
        data = np.reshape(data, (1, y, x))
        block = myokit.DataBlock2d(x, y, [0])
        block.set2d('colormap', data, copy=False)
        return block.images('colormap', colormap=name)[0]

    @staticmethod
    def names():
        """
        Returns an iterator over the names of all available colormaps.
        """
        return ColorMap._colormaps.keys()

    @staticmethod
    def normalize(floats, lower, upper):
        """
        Normalizes the given float data based on the specified lower and upper
        bounds.
        """
        floats = np.array(floats, copy=True)
        # Enforce lower and upper bounds
        floats[floats < lower] = lower
        floats[floats > upper] = upper
        # Normalize
        n = floats - lower
        r = upper - lower
        if r == 0:
            return n
        else:
            return n / r


class ColorMapBlue(ColorMap):
    """
    A nice red colormap.
    """
    def __call__(self, floats, lower=None, upper=None, alpha=True, rgb=None):
        # Normalize floats
        f = ColorMap.normalize(floats, lower, upper)
        # Calculate h,s,v and convert to rgb
        h = np.zeros(f.shape)
        s = f
        v = np.ones(f.shape)
        b, g, r = ColorMap.hsv_to_rgb(h, s, v)
        # Color order
        rgb = (sys.byteorder == 'big') if rgb is None else rgb
        # Offset for first color in (a)rgb or rgb(a)
        m = 1 if (alpha and rgb) else 0
        # Number of bytes per float
        n = 4 if alpha else 3
        # Create output
        out = 255 * np.ones(n * len(floats), dtype=np.uint8)
        out[m + 0::n] = r if rgb else b
        out[m + 1::n] = g
        out[m + 2::n] = b if rgb else r
        return out


class ColorMapGreen(ColorMap):
    """
    A nice green colormap.
    """
    def __call__(self, floats, lower=None, upper=None, alpha=True, rgb=None):
        # Normalize floats
        f = ColorMap.normalize(floats, lower, upper)
        # Calculate h,s,v and convert to rgb
        h = np.zeros(f.shape)
        s = f
        v = np.ones(f.shape)
        g, r, b = ColorMap.hsv_to_rgb(h, s, v)
        # Color order
        rgb = (sys.byteorder == 'big') if rgb is None else rgb
        # Offset for first color in (a)rgb or rgb(a)
        m = 1 if (alpha and rgb) else 0
        # Number of bytes per float
        n = 4 if alpha else 3
        # Create output
        out = 255 * np.ones(n * len(floats), dtype=np.uint8)
        out[m + 0::n] = r if rgb else b
        out[m + 1::n] = g
        out[m + 2::n] = b if rgb else r
        return out


class ColorMapRed(ColorMap):
    """
    A nice red colormap.
    """
    def __call__(self, floats, lower=None, upper=None, alpha=True, rgb=None):
        # Normalize floats
        f = ColorMap.normalize(floats, lower, upper)
        # Calculate h,s,v and convert to rgb
        h = np.zeros(f.shape)
        s = f
        v = np.ones(f.shape)
        r, g, b = ColorMap.hsv_to_rgb(h, s, v)
        # Color order
        rgb = (sys.byteorder == 'big') if rgb is None else rgb
        # Offset for first color in (a)rgb or rgb(a)
        m = 1 if (alpha and rgb) else 0
        # Number of bytes per float
        n = 4 if alpha else 3
        # Create output
        out = 255 * np.ones(n * len(floats), dtype=np.uint8)
        out[m + 0::n] = r if rgb else b
        out[m + 1::n] = g
        out[m + 2::n] = b if rgb else r
        return out


class ColorMapTraditional(ColorMap):
    """
    Traditional hue-cycling colormap.
    """
    def __call__(self, floats, lower=None, upper=None, alpha=True, rgb=None):
        # Normalize floats
        f = ColorMap.normalize(floats, lower, upper)
        # Calculate h,s,v and convert to rgb
        g = 0.6
        s = g + (1 - g) * np.sin(f * 3.14)
        h = (0.85 - 0.85 * f) % 1
        r, g, b = ColorMap.hsv_to_rgb(h, s, s)
        # Color order
        rgb = (sys.byteorder == 'big') if rgb is None else rgb
        # Offset for first color in (a)rgb or rgb(a)
        m = 1 if (alpha and rgb) else 0
        # Number of bytes per float
        n = 4 if alpha else 3
        # Create output
        out = 255 * np.ones(n * len(floats), dtype=np.uint8)
        out[m + 0::n] = r if rgb else b
        out[m + 1::n] = g
        out[m + 2::n] = b if rgb else r
        return out


ColorMap._colormaps['blue'] = ColorMapBlue
ColorMap._colormaps['green'] = ColorMapGreen
ColorMap._colormaps['red'] = ColorMapRed
ColorMap._colormaps['traditional'] = ColorMapTraditional



#
# CellML expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import myokit
from myokit.formats.mathml import MathMLExpressionWriter


class CellMLExpressionWriter(MathMLExpressionWriter):
    """
    Writes equations for variables using CellML's version of MathML.

    Differences from normal MathML:

     1. Only content MathML is supported
     2. Variable names are always written as unames.
     3. Every number defines an attribute cellml:units

    The expression writer requires a single argument ``units``. This should be
    a mapping of Myokit units to string unit names.
    """
    def __init__(self, units=None):
        super(CellMLExpressionWriter, self).__init__()
        super(CellMLExpressionWriter, self).set_mode(presentation=False)
        self._units = {} if units is None else units

    def _ex_number(self, e, t):
        x = self._et.SubElement(t, 'cn')
        x.text = self._fnum(e)
        u = e.unit()
        x.attrib['cellml:units'] = self._units[u] if u else 'dimensionless'

    def _ex_name(self, e, t):
        x = self._et.SubElement(t, 'ci')
        x.text = e.var().uname()

    def _ex_quotient(self, e, t):
        # Note that this _must_ round towards minus infinity!
        # See myokit.Quotient !
        return self.ex(myokit.Floor(myokit.Divide(e[0], e[1])), t)

    def _ex_remainder(self, e, t):
        # Note that this _must_ use the same round-to-neg-inf convention as
        # myokit.Quotient! Implementation below is consistent with Python
        # convention:
        return self.ex(myokit.Minus(
            e[0], myokit.Multiply(e[1], myokit.Quotient(e[0], e[1]))), t)

    def set_mode(self, presentation=False):
        """
        This expression writer only supports content MathML.
        """
        if presentation:
            raise RuntimeError(
                'Presentation MathML is not supported in CellML.')

    def set_lhs_function(self, f):
        """
        This expression writer always uses unames, setting an LHS function is
        not supported.
        """
        raise NotImplementedError



#
# This hidden module contains the GUI elements used throughout Myokit.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

# Library imports
import logging
import os
import platform
import signal
import sys

# Myokit imports
import myokit

# Detect platform
platform = platform.system()

# Select Qt library to use
pyqt4 = False
pyqt5 = False
pyside = False
pyside2 = False

# Allow overriding automatic selection
if myokit.FORCE_PYQT5:
    pyqt5 = True
elif myokit.FORCE_PYQT4:
    pyqt4 = True
elif myokit.FORCE_PYSIDE:
    pyside = True
elif myokit.FORCE_PYSIDE2:
    pyside2 = True
else:
    # Automatic selection
    try:
        import PyQt5  # noqa
        pyqt5 = True
    except ImportError:
        try:
            import PySide2  # noqa
            pyside2 = True
        except ImportError:
            try:
                import PyQt4  # noqa
                pyqt4 = True
            except ImportError:
                try:
                    import PySide  # noqa
                    pyside = True
                except ImportError:
                    raise ImportError('Unable to find PyQt5, PyQt4 or PySide')

# Import and configure Qt
if pyqt5:

    # Load PyQt5
    from PyQt5 import QtGui, QtWidgets, QtCore
    from PyQt5.QtCore import Qt

    # Fix PyQt naming issues
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot
    QtCore.Property = QtCore.pyqtProperty

    # Configure Matplotlib for use with PyQt5
    import matplotlib
    matplotlib.use('Qt5Agg')
    try:
        matplotlib.rcParams['backend.qt5'] = 'PyQt5'
    except KeyError:    # pragma: no cover
        # This is no longer allowed / necessary in matplotlib 3.1.0
        pass
    import matplotlib.backends.backend_qt5agg as matplotlib_backend

    # Set backend variables
    backend = 'PyQt5'
    qtversion = 5

elif pyside2:

    # Load PySide2
    # Load main classes
    from PySide2 import QtGui, QtWidgets, QtCore
    from PySide2.QtCore import Qt

    # Configure Matplotlib for use with PySide2
    import matplotlib
    matplotlib.use('Qt5Agg')
    try:
        matplotlib.rcParams['backend.qt5'] = 'PySide2'
    except KeyError:    # pragma: no cover
        # This is no longer allowed / necessary in matplotlib 3.1.0
        pass
    import matplotlib.backends.backend_qt5agg as matplotlib_backend  # noqa

    # Set backend variables
    backend = 'PySide2'
    qtversion = 5

elif pyqt4:

    # Load PyQt4

    # Deprecated since 2019-09-11
    logger = logging.getLogger('myokit')
    logger.warning(
        'PyQt4 support has been deprecated. Please upgrade to PyQt5 or'
        ' Pyside2.'
    )

    # Set PyQt to "API 2"
    import sip
    sip.setapi('QString', 2)
    sip.setapi('QVariant', 2)
    sip.setapi('QDate', 2)
    sip.setapi('QDateTime', 2)
    sip.setapi('QTextStream', 2)
    sip.setapi('QTime', 2)
    sip.setapi('QUrl', 2)

    # Load main classes
    from PyQt4 import QtGui, QtCore
    from PyQt4.QtCore import Qt

    # Qt5 compatibility
    QtWidgets = QtGui
    QtCore.QItemSelection = QtGui.QItemSelection
    QtCore.QItemSelectionModel = QtGui.QItemSelectionModel
    QtCore.QItemSelectionRange = QtGui.QItemSelectionRange
    QtCore.QSortFilterProxyModel = QtGui.QSortFilterProxyModel
    QtWidgets.QStyleOptionViewItem = QtWidgets.QStyleOptionViewItemV4

    # Fix Qt4 location issue
    import PyQt4.Qt
    QtGui.QKeySequence = PyQt4.Qt.QKeySequence
    QtGui.QTextCursor = PyQt4.Qt.QTextCursor
    del(PyQt4.Qt)

    # Fix PyQt4 naming issues
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot
    QtCore.Property = QtCore.pyqtProperty

    # Fix QFileDialog.getOpenFileName return type issues:
    # Return type in PyQt4 is not a tuple (call ...AndFilter methods instead)
    # Fix getOpenFileName
    def gofn(parent=None, caption='', directory='', filter='',
             initialFilter='', options=0):
        if options == 0:
            options = QtWidgets.QFileDialog.Options()
        return QtWidgets.QFileDialog.getOpenFileNameAndFilter(
            parent, caption, directory, filter, initialFilter, options)
    QtWidgets.QFileDialog.getOpenFileName = staticmethod(gofn)
    del(gofn)

    # Fix getOpenFileNames
    def gofns(parent=None, caption='', directory='', filter='',
              initialFilter='', options=0):
        if options == 0:
            options = QtWidgets.QFileDialog.Options()
        return QtWidgets.QFileDialog.getOpenFileNamesAndFilter(
            parent, caption, directory, filter, initialFilter, options)
    QtWidgets.QFileDialog.getOpenFileNames = staticmethod(gofns)
    del(gofns)

    # Fix getSaveFileName
    def gsfn(parent=None, caption='', directory='', filter='',
             initialFilter='', options=0):
        if options == 0:
            options = QtWidgets.QFileDialog.Options()
        return QtWidgets.QFileDialog.getSaveFileNameAndFilter(
            parent, caption, directory, filter, initialFilter, options)
    QtWidgets.QFileDialog.getSaveFileName = staticmethod(gsfn)
    del(gsfn)

    # Configure Matplotlib for use with PyQt4
    import matplotlib
    matplotlib.use('Qt4Agg')
    try:
        matplotlib.rcParams['backend.qt4'] = 'PyQt4'
    except KeyError:    # pragma: no cover
        # This is no longer allowed / necessary in matplotlib 3.1.0
        pass
    import matplotlib.backends.backend_qt4agg as matplotlib_backend

    # Set backend variables
    backend = 'PyQt4'
    qtversion = 4

elif pyside:

    # Load PySide

    # Deprecated since 2019-09-11
    logger = logging.getLogger('myokit')
    logger.warning(
        'PySide support has been deprecated. Please upgrade to PyQt5 or'
        ' Pyside2.'
    )

    # Load main classes
    from PySide import QtGui, QtCore
    from PySide.QtCore import Qt

    # Qt5 compatibility
    QtWidgets = QtGui
    QtCore.QItemSelection = QtGui.QItemSelection
    QtCore.QItemSelectionModel = QtGui.QItemSelectionModel
    QtCore.QItemSelectionRange = QtGui.QItemSelectionRange
    QtCore.QSortFilterProxyModel = QtGui.QSortFilterProxyModel
    QtWidgets.QStyleOptionViewItem = QtWidgets.QStyleOptionViewItemV4

    # Fix QFileDialog.getOpenFileName signature issues (is different in PySide,
    #  which causes issues when using keyword arguments. This is fixed simply
    #  by simply wrapping the methods.)
    # Signature in PySide
    #  parent
    #  caption
    #  dir --> directory PyQt4/5
    #  filter
    #  selectedFilter --> initialFilter in PyQt4/5
    #  options

    # Fix getOpenFileName
    gofn_org = QtWidgets.QFileDialog.getOpenFileName

    def gofn(parent=None, caption='', directory='', filter='',
             initialFilter='', options=0):
        if options == 0:
            options = QtWidgets.QFileDialog.Options()
        return gofn_org(
            parent, caption, directory, filter, initialFilter, options)
    QtWidgets.QFileDialog.getOpenFileName = staticmethod(gofn)
    del(gofn)

    # Fix getOpenFileNames
    gofns_org = QtWidgets.QFileDialog.getOpenFileNames

    def gofns(parent=None, caption='', directory='', filter='',
              initialFilter='', options=0):
        if options == 0:
            options = QtWidgets.QFileDialog.Options()
        return gofns_org(
            parent, caption, directory, filter, initialFilter, options)
    QtWidgets.QFileDialog.getOpenFileNames = staticmethod(gofns)
    del(gofns)

    # Fix getSaveFileName
    gsfn_org = QtWidgets.QFileDialog.getSaveFileName

    def gsfn(parent=None, caption='', directory='', filter='',
             initialFilter='', options=0):
        if options == 0:
            options = QtWidgets.QFileDialog.Options()
        return gsfn_org(
            parent, caption, directory, filter, initialFilter, options)
    QtWidgets.QFileDialog.getSaveFileName = staticmethod(gsfn)
    del(gsfn)

    # Configure Matplotlib for use with PySide
    import matplotlib
    matplotlib.use('Qt4Agg')
    try:
        matplotlib.rcParams['backend.qt4'] = 'PySide'
    except KeyError:    # pragma: no cover
        # This is no longer allowed / necessary in matplotlib 3.1.0
        pass
    import matplotlib.backends.backend_qt4agg as matplotlib_backend  # noqa

    # Set backend variables
    backend = 'PySide'
    qtversion = 4

else:

    raise Exception('Selection of qt version failed.')

# Delete temporary variables
del(pyqt4, pyqt5, pyside)

# Icons with fallback for apple and windows
ICON_PATH = os.path.join(myokit.DIR_DATA, 'gui')
ICONS = {
    'document-new': 'new.png',
    'document-open': 'open.png',
    'document-save': 'save.png',
    'edit-undo': 'undo.png',
    'edit-redo': 'redo.png',
    'edit-find': 'find.png',
    'media-playback-start': 'run.png',
}
for k, v in ICONS.items():
    ICONS[k] = os.path.join(ICON_PATH, v)

# Toolbar style suitable for platform
TOOL_BUTTON_STYLE = Qt.ToolButtonTextUnderIcon
if platform == 'Windows':   # pragma: no linux cover
    TOOL_BUTTON_STYLE = Qt.ToolButtonIconOnly
elif platform == 'Darwin':  # pragma: no linux cover
    TOOL_BUTTON_STYLE = Qt.ToolButtonTextOnly


# Stand alone applications
class MyokitApplication(QtWidgets.QMainWindow):
    """
    Base class for Myokit applications.

    *Extends*: ``QtWidgets.QMainWindow``.
    """


def icon(name):
    """
    Returns a QtIcon created either from the theme or from one of the fallback
    icons.

    Raises a ``KeyError`` if no such icon is available.
    """
    return QtGui.QIcon.fromTheme(name, QtGui.QIcon(ICONS[name]))


def qtMonospaceFont():
    """
    Attempts to create and return a monospace font.
    """
    font = QtGui.QFont('monospace')
    if platform == 'Windows':   # pragma: no linux cover
        font.setStyleHint(QtGui.QFont.TypeWriter)
    else:
        font.setStyleHint(QtGui.QFont.Monospace)
    font.setHintingPreference(QtGui.QFont.PreferVerticalHinting)  # Qt5
    return font


def run(app, *args):
    """
    Runs a Myokit gui app as a stand-alone application.

    Arguments:

    ``app``
        The application to run, specified as a class object (not an instance).
    ``*args``
        Any arguments to pass to the app's constructor.

    Example usage:

        load(myokit.gui.MyokitIDE, 'model.mmt')


    """
    # Test application class
    if not issubclass(app, MyokitApplication):
        raise ValueError(
            'Application must be specified as a type extending'
            ' MyokitApplication.')

    # Create Qt app
    a = QtWidgets.QApplication([])

    # Apply custom styling if required
    #_style_application(a)
    # Close with last window
    a.lastWindowClosed.connect(a.quit)

    # Close on Ctrl-C
    def int_signal(signum, frame):
        a.closeAllWindows()
    signal.signal(signal.SIGINT, int_signal)

    # Create app and show
    app = app(*args)
    app.show()

    # For some reason, Qt needs focus to handle the SIGINT catching...
    timer = QtCore.QTimer()
    timer.start(500)  # Flags timeout every 500ms
    timer.timeout.connect(lambda: None)

    # Wait for app to exit
    sys.exit(a.exec_())

# def _style_application(app):
#    """
#    Applies some custom styling to a Qt application.
#
#    (Not required if running the app with :meth:`run`)
#    """



#
# Provides CellML support
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

from ._importer import CellMLImporter, CellMLError  # noqa
from ._exporter import CellMLExporter
from ._ewriter import CellMLExpressionWriter

import re


# Importers
_importers = {
    'cellml': CellMLImporter,
}


def importers():
    """
    Returns a dict of all importers available in this module.
    """
    return dict(_importers)


# Exporters
_exporters = {
    'cellml': CellMLExporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'cellml': CellMLExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


# Valid CellML identifiers
# Official docs allow silly things, e.g. 1e2 or 123 or _123
# re.compile('^_*[a-zA-Z0-9][a-zA-Z0-9_]*$')
# So let's be more strict:
#  - At least one letter
#  - Can't start with a number
_cellml_identifier = re.compile('^([_][0-9_]*)?[a-zA-Z][a-zA-Z0-9_]*$')
# Future versions will have:
#   - Must start with at least one letter


def is_valid_identifier(name):
    """
    Tests if the given ``name`` is a valid CellML 1 identifier.
    """
    return _cellml_identifier.match(name) is not None


# Don't expose imported modules
del(re)



#
# Common experiments (activation, inactivation, etc.).
#
# Some functions in this module require a recent version of scipy (i.e. the
# method scipy.optimize.curve_fit must exist).
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import numpy as np
import myokit

# Strings in Python 2 and 3
try:
    basestring
except NameError:   # pragma: no python 2 cover
    basestring = str


#
# Deprecated since 2018-04-16
#
import logging
logger = logging.getLogger('myokit')
logger.warning(
    'The module myokit.lib.common is deprecated: it will be removed in future'
    ' versions of Myokit.'
)


class StepProtocol(object):
    """
    An abstract base class for step protocol experiments.
    """
    def __init__(self, model, var, vvar=None):
        # Clone model
        self._model = model.clone()
        # Get variable names
        self._vars = []
        if isinstance(var, (basestring, myokit.Variable)):
            var = [var]
        for v in var:
            if isinstance(v, myokit.Variable):
                v = v.qname()
            self._vars.append(self._model.get(v).qname())
        # Turn the membrane potential into a constant
        if vvar is None:
            vvar = self._model.label('membrane_potential')
            if vvar is None:
                raise ValueError(
                    'Membrane potential variable must be given as the argument'
                    ' "vvar" or labeled in the model as "membrane_potential".')
        else:
            if isinstance(vvar, myokit.Variable):
                vvar = vvar.qname()
            vvar = self._model.get(vvar)
        if vvar.is_state():
            vvar.demote()
        vvar.set_binding(None)
        vvar.set_rhs(0)
        # Get membrane potential name
        self._vvar = vvar.qname()
        # Get time variable name
        self._tvar = self._model.time().qname()
        # Start without any conversion factor
        self._factor = None
        # Maximum step size during logged simulation or None
        self._max_step_size = None
        # No simulation data yet
        self._logs = None
        # Holding potential & step potentials
        self.set_holding_potential()
        self.set_step_potential()

    def convert_g2i(self, vrev=60, gmax=1):
        """
        Converts any data logged during this experiment from conductances to
        currents by multiplying each step's data by::

            gmax * (v - vrev)

        This can be useful to obtain the original current traces when a
        conductance value was logged or when creating an IV protocol.

        Calling this method will override any previously set conversion factor.
        """
        self._factor = float(gmax) * (self._steps - float(vrev))

    def convert_i2g(self, vrev=60, gmax=1):
        """
        Converts any data logged during this experiment from currents to
        conductance by multiplying each step's data by::

            1 / (gmax * (v - vrev))

        This can be useful to obtain an activation or inactivation curve from
        logged current data. However, since this leads to numerical issues
        around ``v == vrev`` its better to run these experiments directly on a
        conductance variable.

        Calling this method will override any previously set conversion factor.
        """
        self._factor = 1.0 / (float(gmax) * (self._steps - float(vrev)))

    def disable_conversion(self):
        """
        Disables any previously set conversion factor.
        """
        self._factor = None

    def fit_boltzmann(self, var=None):
        """
        Attempts to fit a Boltzmann curve to a voltage-peaks dataset.

        The variable whose peaks to use can be specified as ``var``. If no
        variable is given the first value specified in the constructor is used.
        The fitted curve is given by::

            g = gmin + (gmax - gmin) / (1 + np.exp((v - v_half) / k))

        Note: This method requires a scipy installation providing the method
        scipy.optimize.curve_fit
        """
        from scipy.optimize import curve_fit
        d = self.peaks(normalize=True).npview()
        v = d[self._vvar]
        g = d[var] if var else d[self._vars[0]]
        o = np.ones(v.shape)
        gmin = np.min(g)
        gmax = np.max(g)

        def f(v, v_half, k):
            return np.select(
                [v == v_half],
                [o * k],
                gmin + (gmax - gmin) / (1.0 + np.exp((v - v_half) / k)))

        vmid = v[-1] - v[0]
        slope = -5 if g[0] < g[-1] else 5
        p = curve_fit(f, v, g, [vmid, slope])
        return p[0]

    def peaks(self, normalize=False):
        """
        Returns a :class:`myokit.DataLog` containing the tested step
        voltages and the peak values of the logged variable(s).

        The names used in the simulation log correspond to those used in the
        model. For example, when doing an experiment on a Sodium channel the
        simulation log might have entries ``membrane.v`` and ``ina.g`` where
        ``membrane.v`` contains the used voltage steps while ``ina.g`` contains
        the peak values measured at those voltages.

        If ``normalize`` is set to ``True``, the peak data returned will be
        normalized by dividing all values by the largest (most positive) peak
        in the list. If no positive, non-zero values are found no normalization
        will be applied.

        If any conversion factor was specified the data will be converted
        before normalization.
        """
        # Run simulation if needed
        if self._logs is None:
            self._run()
        # Create a copy of the voltage steps
        v = np.array(self._steps, copy=True)
        # Create a simulation log
        d = myokit.DataLog()
        d[self._vvar] = v
        # Factor
        factor = self._factor if self._factor is not None else 1
        # Find the peaks
        for var in self._vars:
            peaks = np.zeros(len(v))
            for k, log in enumerate(self._logs):
                peaks[k] = log[var][np.argmax(np.abs(log[var]))]
            d[var] = peaks * factor
        # Normalize (only if log contains positive values)
        if normalize:
            for var in self._vars:
                x = d[var]
                m = np.max(x)
                if m > 0:
                    d[var] = x / m
        return d

    def _run(self):
        """
        Should run the simulation and save the current traces in self._logs
        """
        raise NotImplementedError

    def set_constant(self, var, value):
        """
        Changes the value of a constant in the used model.
        """
        if isinstance(var, myokit.Variable):
            var = var.qname()
        var = self._model.get(var)
        self._model.set_value(var, value)
        self._logs = None

    def set_holding_potential(self, vhold=-140, thold=1800):
        """
        Sets the holding potential and the time to hold. During the experiment,
        the cell will be held at ``vhold`` for ``thold`` time units before
        every voltage step.
        """
        vhold = float(vhold)
        thold = float(thold)
        if thold < 0:
            raise ValueError('Time to hold cannot be negative.')
        self._vhold = vhold
        self._thold = thold
        self._logs = None

    def set_max_step_size(self, dtmax=None):
        """
        Can be used to set a maximum step size to use in the logged parts of
        the simulation. Use ``dtmax==None`` to let the solver chose any size it
        likes.
        """
        self._max_step_size = None if dtmax is None else float(dtmax)

    def set_step_potential(self, vmin=-100, vmax=50, dv=1, tstep=200):
        """
        Sets the step potentials and the step duration. Each experiment will
        step linearly from ``vmin`` to ``vmax`` with steps of size ``dv``. The
        cell is held at each step for ``tstep`` time units.
        """
        vmin = float(vmin)
        vmax = float(vmax)
        dv = float(dv)
        tstep = float(tstep)
        if vmax <= vmin:
            raise ValueError(
                'Maximum voltage must be greater than minimum voltage.')
        if dv <= 0:
            raise ValueError(
                'The voltage increment dv must be greater than zero.')
        if tstep < 0:
            raise ValueError('Step duration cannot be negative.')
        self._vmin = vmin
        self._vmax = vmax
        self._dv = dv
        self._tstep = tstep
        self._steps = np.arange(self._vmin, self._vmax + self._dv, self._dv)
        self._logs = None

    def steps(self):
        """
        Returns the list of steps this protocol will use.
        """
        return list(self._steps)

    def times(self):
        """
        Returns a :class:`myokit.DataLog` containing the time-to-peak for
        each logged variable at each voltage step.
        """
        # Run simulation if needed
        if self._logs is None:
            self._run()
        # Create a copy of the voltage steps
        v = np.array(self._steps, copy=True)
        # Create a simulation log
        d = myokit.DataLog()
        d[self._vvar] = v
        # Find the peaks
        for var in self._vars:
            times = np.zeros(len(v))
            for k, log in enumerate(self._logs):
                times[k] = log[self._tvar][np.argmax(np.abs(log[var]))]
            d[var] = times
        return d

    def traces(self):
        """
        Returns the logged traces for each variable as an ordered list of
        tuples ``(v, DataLog)``.

        If any conversion factor was specified the data will be converted
        before returning.
        """
        if self._logs is None:
            self._run()
        data = []
        steps = iter(self._steps)
        factor = self._factor
        if factor is None:
            factor = np.ones(len(self._steps))
        for k, log in enumerate(self._logs):
            v = next(steps)
            d = myokit.DataLog()
            for var in self._vars:
                d[var] = np.array(log[var]) * factor[k]
                d[self._tvar] = log[self._tvar]
            data.append((v, d))
        return data


class Activation(StepProtocol):
    """
    Runs a step protocol and measures during the step. Can be used to create
    activation curves and I-V curves.

    The protocol is defined as follows: Initially, the membrane potential is
    held at a holding potential ``vhold`` for the duration of ``thold`` time
    units. Then, the voltage is changed to ``vstep`` and held there for
    ``vstep`` time units. During this step, the cell's response is logged. The
    experiment is repeated for ``vstep`` values ranging linearly from ``vmin``
    to ``vmax`` with an increment of ``dv``.

    ::

      .               +--- vstep ---+
      .               +-------------+
      .               +-------------+
      .               |             |
      .  vhold -------+             +-
      .  t=0          t=thold       t=thold+tstep
      .  no current   current!

    Accepts the following input arguments:

    ``model``
        The model for which to run the simulations
    ``var``
        A variable or a list of variables to log and use in calculations
    ``vvar``
        The membrane potential variable or its qname. If not given, the method
        will search for a variable labelled ``membrane_potential``.

    Depending on the experiment being run, ``var`` could be a conductance or a
    current variable (or a list of conductances or currents).

    The experiment is not performed until a call to one of the post-processing
    methods is made. After this, the raw data will be cached. Any change to the
    protocol variables after this point will delete the cached data.
    """
    def __init__(self, model, var, vvar=None):
        super(Activation, self).__init__(model, var, vvar)

    def _run(self):
        """
        Runs the experiment, logs during the voltage steps.
        """
        log = self._vars + [self._tvar]
        s = myokit.Simulation(self._model)
        d = []
        for v in self._steps:
            s.reset()
            s.set_constant(self._vvar, self._vhold)
            s.set_max_step_size(None)
            s.pre(self._thold)
            s.set_constant(self._vvar, v)
            s.set_max_step_size(self._max_step_size)
            d.append(s.run(self._tstep, log=log))
        self._logs = d


class Inactivation(StepProtocol):
    """
    Can run an inactivation step protocol on a model and calculate various
    entities using the results.

    The protocol starts by holding the membrane potential at a high value,
    causing the channels to fully activate and then inactivate. Next, the
    potential is stepped to a lower value, causing the inactivation to go away
    while the activation stays close to full. Current is measured after the
    step, when the cell is at the holding potential.

    ::

      .  --- vstep ---+
      .  -------------+
      .  -------------+
      .               |
      .               +--- vhold ---+-
      .  t=0          t=tstep       t=tstep+thold

    Accepts the following input arguments:

    ``model``
        The model for which to run the simulations
    ``var``
        A variable or a list of variables to log and use in calculations
    ``vvar``
        The membrane potential variable or its qname. If not given, the method
        will search for a variable labelled ``membrane_potential``.

    Depending on the experiment being run, ``var`` could be a conductance or a
    current variable (or a list of conductances or currents).

    The experiment is not performed until a call to one of the post-processing
    methods is made. After this, the raw data will be cached. Any change to the
    protocol variables after this point will delete the cached data.
    """
    def __init__(self, model, var, vvar=None):
        super(Inactivation, self).__init__(model, var, vvar)

    def _run(self):
        """
        Runs the simulation, saves the current traces.
        """
        log = self._vars + [self._tvar]
        d = []
        s = myokit.Simulation(self._model)
        for v in self._steps:
            s.reset()
            s.set_constant(self._vvar, v)
            s.pre(self._tstep)
            s.set_constant(self._vvar, self._vhold)
            d.append(s.run(self._thold, log=log))
        self._logs = d

    def set_holding_potential(self, vhold=-20, thold=50):
        """
        Sets the holding potential and the time to hold. During the experiment,
        the cell will be held at ``vhold`` for ``thold`` time units before
        every voltage step.
        """
        super(Inactivation, self).set_holding_potential(vhold, thold)

    def set_step_potential(self, vmin=-100, vmax=-40, dv=1, tstep=1000):
        """
        Sets the step potentials and the step duration. Each experiment will
        step linearly from ``vmin`` to ``vmax`` with steps of size ``dv``. The
        cell is held at each step for ``tstep`` time units.
        """
        super(Inactivation, self).set_step_potential(vmin, vmax, dv, tstep)


class Recovery(object):
    """
    Can run a two-pulse recovery from inactivation experiment and process the
    results.

    The experiment proceeds as follows: first, the cell is held at a low
    holding potential ``vhold`` for ``thold`` time units. This causes the
    channels to fully deactivate (activation=0) and fully recover from
    inactivation (inactivation=1). Next, the membrane is stepped up to a
    voltage ``vstep`` and kept there for ``tstep1`` time units. This causes the
    channels to activate (activation=1) and then inactivate (inactivation=0).
    The model then steps back down to ``vhold`` for ``twait`` time units,
    during which partial recovery from inactivation is expected. After this
    short recovery period, the cell is stepped back to ``vstep`` for another
    ``tstep2`` time units. The recovery from inactivation can then be judged
    by comparing the peak current during step2 with the peak current during
    step1. By varying the time between steps ``twait`` a plot of recovery
    characteristics can be made.

    ::

      .                +--- vstep ---+         +- vstep -+
      .                |             |  twait  |         |
      .                |             | <-----> |         |
      .                |             |         |         |
      .  +--- vhold ---+             +- vhold -+         +---
      .  t=0           t=thold       t+=tstep1 t+=twait  t+=tstep2

    Accepts the following input arguments:

    ``model``
        The model for which to run the simulations
    ``var``
        A conductance variable (or a list of variables) to record during the
        experiment. Variables can be specified using Variable objects or
        through their fully qualified names.
    ``vvar``
        The membrane potential variable or its qname. If not given, the method
        will search for a variable labelled ``membrane_potential``.

    The argument ``var`` expects conduction variables, not currents. In other
    words, it expects the name of a variable ``g`` such that
    ``I = Gmax * g * (V - E)``. Using conductance in this test rather than
    current avoids the numerical problems incurred by dividing ``I`` through
    ``(V-E)``.
    """
    def __init__(self, model, var, vvar=None):
        if not model.is_valid():
            raise ValueError('This method requires a valid model.')
        # Clone model & variables
        self._model = model.clone()
        # Get time variable
        self._tvar = self._model.time()
        # Check conductance variables
        self._vars = []
        if isinstance(var, (basestring, myokit.Variable)):
            var = [var]
        for v in var:
            if isinstance(v, myokit.Variable):
                v = v.qname()
            elif '.' not in v:
                raise ValueError(
                    'The variable name(s) given as var must be given as fully'
                    ' qualified names <component.variable>.')
            self._vars.append(self._model.get(v))
        # Check membrane potential
        if vvar is None:
            self._vvar = self._model.label('membrane_potential')
            if self._vvar is None:
                raise ValueError(
                    'Membrane potential variable must be given by vvar or'
                    ' specified using the label "membrane_potential".')
        else:
            if isinstance(vvar, myokit.Variable):
                vvar = vvar.qname()
            elif '.' not in vvar:
                raise ValueError(
                    'The variable name vvar must be given as a fully qualified'
                    ' variable name <component.var>.')
            self._vvar = self._model.get(vvar)
        # Update membrane potential variable
        if self._vvar.is_state():
            self._vvar.demote()
        self._vvar.set_binding(None)
        # Set voltages
        self.set_holding_potential()
        self.set_step_potential()
        self.set_pause_duration()

    def ratio(self):
        """
        Returns the ratios of the peak conductances (p1 / p2) for step 1 and
        step 2.

        The returned value is a :class:`myokit.DataLog` with entries
        corresponding to the given variable names.
        """
        # Times to wait
        twaits = np.exp(
            np.linspace(np.log(self._tmin), np.log(self._tmax), self._nt))
        # Variables to log
        log_vars = [x.qname() for x in self._vars]
        # Run simulations
        self._vvar.set_rhs(self._vhold)     # Make V a constant
        s = myokit.Simulation(self._model)
        log = myokit.DataLog()
        gvars = [x.qname() for x in self._vars]
        for g in gvars:
            log[g] = []
        log[self._tvar.qname()] = list(twaits)
        for twait in twaits:
            s.set_constant(self._vvar, self._vhold)
            s.run(self._thold, log=myokit.LOG_NONE)
            s.set_constant(self._vvar, self._vstep)
            d1 = s.run(self._tstep1, log=log_vars)
            s.set_constant(self._vvar, self._vhold)
            s.run(twait, log=myokit.LOG_NONE)
            s.set_constant(self._vvar, self._vstep)
            d2 = s.run(self._tstep2, log=log_vars)
            for g in gvars:
                ratio = np.max(d1[g])
                ratio = np.nan if ratio == 0 else np.max(d2[g]) / ratio
                log[g].append(ratio)
        return log

    def set_holding_potential(self, vhold=-120, thold=2000):
        """
        Sets the holding potential and the time to hold. During the experiment,
        the cell will be held at ``vhold`` for ``thold`` time units before
        every voltage step.
        """
        vhold = float(vhold)
        thold = float(thold)
        if thold < 0:
            raise ValueError('Time to hold cannot be negative.')
        self._vhold = vhold
        self._thold = thold

    def set_pause_duration(self, tmin=0.5, tmax=1000, nt=50):
        """
        Sets the duration of the pauses between the steps.

        ``tmin``
            The shortest time between steps
        ``tmax``
            The longest time between steps
        ``nt``
            The number of times to test

        """
        tmin = float(tmin)
        tmax = float(tmax)
        nt = int(nt)
        if tmin < 0:
            raise ValueError('Minimum time cannot be negative.')
        if tmax < 0:
            raise ValueError('Maximum time cannot be positive.')
        if tmin >= tmax:
            raise ValueError('Maximum time must be grater than minimum.')
        if nt < 2:
            raise ValueError(
                'The number of times to test must be greater than one.')
        self._tmin = tmin
        self._tmax = tmax
        self._nt = nt

    def set_step_potential(self, vstep=-20, tstep1=500, tstep2=25):
        """
        Sets the step potential and the step durations.

        ``vstep``
            The potential used during the steps
        ``tstep1``
            The duration of the first pulse
        ``tstep2``
            The duration of the second pulse

        """
        vstep = float(vstep)
        tstep1 = float(tstep1)
        tstep2 = float(tstep2)
        if tstep1 < 0:
            raise ValueError('Time of first step cannot be negative.')
        if tstep2 < 0:
            raise ValueError('Time of second step cannot be negative.')
        self._vstep = vstep
        self._tstep1 = tstep1
        self._tstep2 = tstep2


class Restitution(object):
    """
    Can run a restitution experiment and return the values needed to make a
    plot.

    Accepts the following input arguments:

    ``model``
        The model for which to run the simulations
    ``vvar``
        The variable or variable name representing membrane potential. If not
        given, the method will look for the label ``membrane_potential``, if
        that's not found an exception is raised.

    """
    def __init__(self, model, vvar=None):
        # Check model
        self._model = model.clone()

        # Check membrane potential
        if vvar is None:
            self._vvar = self._model.label('membrane_potential')
            if self._vvar is None:
                raise ValueError(
                    'Membrane potential variable must be given by vvar or'
                    ' specified using the label "membrane_potential".')
        else:
            if isinstance(vvar, myokit.Variable):
                vvar = vvar.qname()
            elif '.' not in vvar:
                raise ValueError(
                    'The variable name vvar must be given as a fully qualified'
                    ' variable name <component.var>.')
            self._vvar = self._model.get(vvar)

        # Set default arguments
        self.set_max_step_size()
        self.set_times()
        self.set_beats()
        self.set_stimulus()
        self.set_threshold()

        # No data yet!
        self._data = None

    def _run(self):
        """
        Runs the simulations, saves the data.
        """
        # Create protocol
        e = {
            'level': self._stim_level,
            'start': 0,
            'duration': self._stim_duration,
            'period': self._clmin,
            'multiplier': 0,
        }

        # Create simulation
        s = myokit.Simulation(self._model, apd_var=self._vvar)
        s.set_max_step_size(self._max_step_size)

        # Start testing
        i = 0
        pcls = []
        apds = []
        c = self._clmax
        while c >= self._clmin:
            # Update cycle length
            c = self._clmax - i * self._dcl
            i += 1

            # Create and set new protocol
            p = myokit.Protocol()
            e['period'] = c
            p.schedule(**e)
            s.set_protocol(p)

            # Run simulation
            s.reset()
            s.pre(c * self._pre_beats)
            d, a = s.run(
                c * self._beats,
                log=myokit.LOG_NONE,
                apd_threshold=self._apd_threshold
            )

            # Save apds
            for apd in a['duration']:
                pcls.append(c)
                apds.append(apd)

        # Store data
        self._data = pcls, apds

    def run(self):
        """
        Returns a :class:`DataLog` containing the tested cycle lengths as
        ``cl`` and the measured action potential durations as ``apd``. The
        diastolic intervals are given as ``di``.

        Each cycle length is repeated ``beats`` number of times, where
        ``beats`` is the number of beats specified in the constructor.
        """
        # Run
        if self._data is None:
            self._run()
        # Get data
        cl, apd = self._data
        d = myokit.DataLog()
        d['cl'] = list(cl)
        d['apd'] = list(apd)
        d['di'] = list(np.array(cl, copy=False) - np.array(apd, copy=False))
        return d

    def set_beats(self, beats=2, pre=50):
        """
        Sets the number of beats each cycle length is tested for.

        ``beats``
            The number of beats during which apd is measured
        ``pre``
            The number of pre-pacing beats done at each cycle length before the
            measurement.
        """
        beats = int(beats)
        pre = int(pre)
        if beats < 1:
            raise ValueError(
                'The number of beats must be an integer greater than zero.')
        if pre < 0:
            raise ValueError(
                'The number of pre-pacing beats must be a positive integer.')
        self._beats = beats
        self._pre_beats = pre
        self._data = None

    def set_max_step_size(self, dtmax=None):
        """
        Sets an (optional) maximum step size for the solver. To let the solver
        pick any step size it likes, use ``dtmax=None``.

        This method can be useful to avoid "CVODE flag 22" errors.
        """
        if dtmax is None:
            self._max_step_size = None
        else:
            dtmax = float(dtmax)
            if dtmax <= 0:
                raise ValueError(
                    'Maximum step size must be greater than zero.')
            self._max_step_size = dtmax
        self._data = None

    def set_stimulus(self, duration=2.0, level=1):
        """
        Sets the stimulus used to pace the model.

        ``stim_duration``
            The duration of the pacing stimulus.
        ``stim_level``
            The level of the dimensionless pacing stimulus.

        """
        duration = float(duration)
        level = float(level)
        if duration < 0:
            raise ValueError('The duratio cannot be negative.')
        self._stim_duration = duration
        self._stim_level = level
        self._data = None

    def set_threshold(self, threshold=-70):
        """
        Sets the APD threshold, specified as a fixed membrane potential.
        """
        self._apd_threshold = float(threshold)
        self._data = None

    def set_times(self, clmin=300, clmax=1200, dcl=20):
        """
        Sets the pacing cycle lengths tested in this experiment.

        ``clmin``
            The shortest cycle length tested.
        ``clmax``
            The longest cycle length tested.
        ``dcl``
            The size of the steps from ``clmin`` to ``clmax``

        """
        clmin = float(clmin)
        clmax = float(clmax)
        dcl = float(dcl)
        if clmin < 0:
            raise ValueError('Minimum time cannot be negative.')
        if clmin >= clmax:
            raise ValueError('Minimum time must be smaller than maximum time.')
        if dcl <= 0:
            raise ValueError('Step size must be greater than zero')
        self._clmin = clmin
        self._clmax = clmax
        self._dcl = dcl
        self._data = None


class StrengthDuration(object):
    """
    For a range of durations, this experiment checks the stimulus size needed
    to excite a cell.

    Accepts the following input arguments:

    ``model``
        The model to run the experiment with.
    ``ivar``
        A variable that can be used as the stimulus current.  It's value will
        be set as ``pace * amplitude`` where pace is a variable bound to the
        pacing signal and amplitude is varied by the method.
    ``vvar``
        A variable that indicates the membrane potential. If not specified (or
        given as ``None``), the method will search for a variable labeled as
        ``membrane_potential``.

    """
    def __init__(self, model, ivar, vvar=None):
        # Clone model
        self._model = model.clone()
        # Get stimulus current variable
        self._ivar = self._model.get(ivar)

        # Get membrane potential variable
        if vvar is None:
            self._vvar = model.label('membrane_potential')
            if self._vvar is None:
                raise ValueError(
                    'This method requires the membrane potential variable to'
                    ' be passed in as `vvar` or indicated in the model using'
                    ' the label `membrane_potential`.')
        else:
            if isinstance(vvar, myokit.Variable):
                vvar = vvar.qname()
            self._vvar = model.get(vvar)

        # Get time variable
        self._tvar = self._model.time()
        del(model, vvar, ivar)

        # Unbind any existing pace variable
        var = self._model.binding('pace')
        if var is not None:
            var.set_binding(None)

        # Create new pacing variable
        c = self._ivar.parent(myokit.Component)

        def add_variable(c, name):
            try:
                return c.add_variable(name)
            except myokit.DuplicateName:
                i = 2
                n = name + str(i)
                while True:
                    try:
                        return c.add_variable(n)
                    except myokit.DuplicateName:
                        i += 1
                        n = name + str(i)
        self._pvar = add_variable(c, 'pace')
        self._pvar.set_binding('pace')
        self._pvar.set_rhs(0)

        # Create new amplitude variable
        self._avar = add_variable(c, 'amplitude')
        self._avar.set_rhs(0)

        # Set rhs of current variable
        if self._ivar.is_state():
            self._ivar.demote()
        self._ivar.set_rhs(myokit.Multiply(self._pvar.lhs(), self._avar.lhs()))

        # Set default parameters
        self.set_currents()
        self.set_precision()
        self.set_threshold()
        self.set_times()

        # No data yet!
        self._data = None

    def run(self, debug=False):
        """
        Runs the experiment, returns a :class:`myokit.DataLog` with the
        entries `durations` and `amplitudes`, where the value in `amplitudes`
        is the minimum strength required to create a depolarization at the
        corresponding duration.
        """
        if self._data is None:
            self._run(debug)
        return self._data

    def _run(self, debug=False):
        """
        Inner version of run()
        """
        if debug:
            import traceback
        # Create simulation
        s = myokit.Simulation(self._model)
        # Variables to log
        vvar = self._vvar.qname()
        # Output data
        durations = np.array(self._durations, copy=True)
        amplitudes = np.zeros(durations.shape)
        # Test every duration
        for k, duration in enumerate(durations):
            if debug:
                print('Testing duration: ' + str(duration))
            s.set_protocol(myokit.pacing.blocktrain(self._time + 1, duration))
            a1 = self._amin
            a2 = self._amax
            # Test minimum amplitude
            s.reset()
            s.set_constant(self._avar, a1)
            try:
                d = s.run(self._time, log=[vvar]).npview()
                t1 = (np.max(d[vvar]) > self._threshold)
            except Exception:
                if debug:
                    traceback.print_exc()
                t1 = False
            if debug:
                print(t1)
            # Test maximum amplitude
            s.reset()
            s.set_constant(self._avar, a2)
            try:
                d = s.run(self._time, log=[vvar]).npview()
                t2 = (np.max(d[vvar]) > self._threshold)
            except Exception:
                if debug:
                    traceback.print_exc()
                t2 = False
            if debug:
                print(t2)
            if t1 == t2:
                # No zero crossing found
                amplitudes[k] = np.nan
                if debug:
                    print('> no zero crossing')
                continue
            # Zero must lie in between. Start bisection search
            a = 0.5 * a1 + 0.5 * a2
            for j in range(0, self._precision):
                s.reset()
                s.set_constant(self._avar, a)
                try:
                    d = s.run(self._time, log=[vvar]).npview()
                except Exception:
                    if debug:
                        traceback.print_exc()
                    break
                t = (np.max(d[vvar]) > self._threshold)
                if t1 == t:
                    a1 = a
                else:
                    a2 = a
                a = 0.5 * a1 + 0.5 * a2
            amplitudes[k] = a
            if debug:
                print('> ' + str(a))
        # Set output data
        self._data = myokit.DataLog()
        self._data['duration'] = durations
        self._data['strength'] = amplitudes

    def set_currents(self, imin=-250, imax=0):
        """
        Sets the range of current levels tested.

        ``imin``
            The lowest (or most negative) current amplitude to test.
        ``imax``
            The greatest (or least negative) current amplitude to test.
        """
        amin = float(imin)
        amax = float(imax)
        if amin >= amax:
            raise ValueError(
                'Minimum current amplitude must be smaller than maximum'
                ' value.')
        self._amin = amin
        self._amax = amax
        self._data = None

    def set_precision(self, precision=10):
        """
        Sets the number of different amplitudes tried. This is done using a
        bisection algorithm, so low values of ``precision`` can still produce a
        good result.
        """
        precision = int(precision)
        if precision < 1:
            raise ValueError('The number of tries must be greater than zero.')
        self._precision = precision
        self._data = None

    def set_threshold(self, threshold=10):
        """
        Sets the level above which the membrane potential must rise to count as
        a depolarization.
        """
        self._threshold = float(threshold)
        self._data = None

    def set_times(self, tmin=0.2, tmax=2.0, dt=0.1, twait=50):
        """
        Sets the tested stimulus durations.

        ``tmin``
            The smallest stimulus time tested.
        ``tmax``
            The largest stimulus time tested.
        ``dt``
            The step size when going from tmin to tmax.
        ``twait``
            The duration of the experiment that tries to measure a
            depolarization.

        """
        tmin = float(tmin)
        tmax = float(tmax)
        if tmin < 0:
            raise ValueError('Minimum time cannot be negative.')
        if tmax < 0:
            raise ValueError('Maximum time cannot be negative.')
        if tmin >= tmax:
            raise ValueError(
                'The maximum time must be greater than the minimum time.')
        dt = float(dt)
        if dt <= 0:
            raise ValueError('The step size must be greater than zero.')
        twait = float(twait)
        if twait < 0:
            raise ValueError('The time "twait" must be greater than zero.')
        self._durations = np.arange(tmin, tmax, dt)
        self._time = twait
        self._data = None



#
# Exports to CellML
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import xml.etree.cElementTree as et

import myokit
import myokit.units
import myokit.formats


# Quoting URI strings in Python2 and Python3
try:
    from urllib.parse import quote
except ImportError:     # pragma: no python 3 cover
    # Python 2
    from urllib import quote


# Namespaces
NS_BQBIOL = 'http://biomodels.net/biology-qualifiers/'
NS_CELLML_1_0 = 'http://www.cellml.org/cellml/1.0#'
NS_CMETA = 'http://www.cellml.org/metadata/1.0#'
NS_MATHML = 'http://www.w3.org/1998/Math/MathML'
NS_OXMETA = 'https://chaste.comlab.ox.ac.uk/cellml/ns/oxford-metadata#'
NS_RDF = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
#NS_RDFS = 'http://www.w3.org/2000/01/rdf-schema#'
NS_TMP_DOC = 'http://cellml.org/tmp-documentation'


class CellMLExporter(myokit.formats.Exporter):
    """
    This:class:`Exporter <myokit.formats.Exporter>` creates a CellML model.
    """
    def __init__(self):
        super(CellMLExporter, self).__init__()

    def custom_unit_name(self, unit):
        """
        Creates an almost readable name for a custom Myokit unit.
        """
        import myokit.formats.cellml as cellml

        # Get preferred name
        name = str(unit)[1:-1]

        # Check if that's allowed
        if cellml.is_valid_identifier(name):
            return name

        # Create custom name

        # Split unit from multiplier part
        if ' ' in name:
            name, multiplier = name.split(' ')
        else:
            name, multiplier = name, ''

        # Treat unit parts
        name = name.replace('^', '')
        name = name.replace('/', '_per_')
        name = name.replace('*', '_')
        if name[:2] == '1_':
            # E.g. [1_per_mV]
            name = name[2:]
        elif name == '1':
            # E.g. [1 (1000)]
            name = 'dimensionless'

        # Treat multiplier
        if multiplier:
            multiplier = unit.multiplier_log_10()

            # If nice round int, then use e-notation
            if myokit._feq(multiplier, int(multiplier)):
                multiplier = '1e' + str(int(multiplier))
            else:
                multiplier = str(unit.multiplier())

            # Remove characters not allowed in CellML identifiers
            multiplier = multiplier.replace('+', '')
            multiplier = multiplier.replace('-', '_minus_')
            multiplier = multiplier.replace('.', '_dot_')
            name += '_times_' + multiplier

        return name

    def info(self):
        import inspect
        return inspect.getdoc(self)

    def model(
            self, path, model, protocol=None, add_hardcoded_pacing=True,
            pretty_xml=True):
        """
        Writes a CellML model to the given filename.

        Arguments:

        ``path``
            The path/filename to write the generated code too.
        ``model``
            The model to export
        ``protocol``
            This argument will be ignored: protocols are not supported by
            CellML.
        ``add_hardcoded_pacing``
            Set this to ``True`` to add a hardcoded pacing signal to the model
            file. This requires the model to have a variable bound to `pace`.
        ``pretty_xml``
            Set this to ``True`` to write the output in formatted "pretty"
            xml.

        Notes about CellML export:

        * CellML expects a unit for every number present in the model. Since
          Myokit allows but does not enforce this, the resulting CellML file
          may only validate with unit checking disabled.
        * Files downloaded from the CellML repository typically have a pacing
          stimulus embedded in them, while Myokit views models and pacing
          protocols as separate things. To generate a model file with a simple
          embbeded protocol, add the optional argument
          ``add_hardcoded_pacing=True``.
        * Variables annotated with an ``oxmeta`` property will be annotated
          using the oxmeta namespace in the created CellML. For example, a
          variable with the meta-data ``oxmeta: time`` will be annotated as
          ``https://chaste.comlab.ox.ac.uk/cellml/ns/oxford-metadata#time`` in
          the CellML file.

        """
        import myokit.formats.cellml as cellml

        path = os.path.abspath(os.path.expanduser(path))

        # Clear log
        self.logger().clear()
        self.logger().clear_warnings()

        # Replace the pacing variable with a hardcoded stimulus protocol
        if add_hardcoded_pacing:

            # Check for pacing variable
            if model.binding('pace') is None:
                self.logger().warn(
                    'No variable bound to "pace", unable to add hardcoded'
                    ' stimulus protocol.')
            else:
                # Clone model before making changes
                model = model.clone()

                # Get pacing variable
                pace = model.binding('pace')

                # Set basic properties for pace
                pace.set_unit(myokit.units.dimensionless)
                pace.set_rhs(0)
                pace.set_binding(None)
                pace.set_label(None)    # Should already be true...

                # Get time variable of cloned model
                time = model.time()

                # Get time unit
                time_unit = time.unit(mode=myokit.UNIT_STRICT)

                # Get correction factor if using anything other than
                # milliseconds (hardcoded below)
                try:
                    time_factor = myokit.Unit.conversion_factor(
                        'ms', time_unit)
                except myokit.IncompatibleUnitError:
                    time_factor = 1

                # Create new component for the pacing variables
                component = 'stimulus'
                if model.has_component(component):
                    root = component
                    number = 1
                    while model.has_component(component):
                        number += 1
                        component = root + '_' + str(number)
                component = model.add_component(component)

                # Move pace. This will be ok any references: since pace was
                # bound it cannot be a nested variable.
                # While moving, update its name to avoid conflicts with the
                # hardcoded names.
                pace.parent().move_variable(pace, component, new_name='pace')

                # Add variables defining pacing protocol
                qperiod = myokit.Quantity('1000 [ms]')
                qoffset = myokit.Quantity('100 [ms]')
                qduration = myokit.Quantity('2 [ms]')
                period = component.add_variable('period')
                period.set_unit(time_unit)
                period.set_rhs(qperiod * time_factor)
                offset = component.add_variable('offset')
                offset.set_unit(time_unit)
                offset.set_rhs(qoffset * time_factor)
                duration = component.add_variable('duration')
                duration.set_unit(time_unit)
                duration.set_rhs(qduration * time_factor)

                # Add corrected time variable
                ctime = component.add_variable('ctime')
                ctime.set_unit(time_unit)
                ctime.set_rhs(
                    time.qname() + ' - floor(' + time.qname()
                    + ' / period) * period')

                # Remove any child variables pace might have before changing
                # its RHS (which needs to refer to them).
                pace_kids = list(pace.variables())
                for kid in pace_kids:
                    pace.remove_variable(kid, recursive=True)

                # Set new RHS for pace
                pace.set_rhs(
                    'if(ctime >= offset and ctime < offset + duration, 1, 0)')

        # Validate model
        model.validate()

        # Get time variable
        time = model.time()

        # Create model xml element
        emodel = et.Element('model')
        emodel.attrib['xmlns'] = NS_CELLML_1_0
        emodel.attrib['xmlns:cellml'] = NS_CELLML_1_0

        # Add name in 'tmp-documentation' format
        emodel.attrib['name'] = 'generated_model'
        if 'name' in model.meta:
            dtag = et.SubElement(emodel, 'documentation')
            dtag.attrib['xmlns'] = NS_TMP_DOC
            atag = et.SubElement(dtag, 'article')
            ttag = et.SubElement(atag, 'title')
            ttag.text = model.meta['name']

        # Add custom units, create unit map
        exp_si = [si_units[x] for x in myokit.Unit.list_exponents()]
        unit_map = {}   # Add si units later

        def add_unit(unit):
            """
            Checks if the given unit needs to be added to the list of custom
            units and adds it if necessary.
            """
            # Check if already defined
            if unit is None or unit in unit_map or unit in si_units:
                return

            # Create unit name
            name = self.custom_unit_name(unit)

            # Create unit tag
            utag = et.SubElement(emodel, 'units')
            utag.attrib['name'] = name

            # Add part for each of the 7 SI units
            m = unit.multiplier()
            for k, e in enumerate(unit.exponents()):
                if e != 0:
                    tag = et.SubElement(utag, 'unit')
                    tag.attrib['units'] = exp_si[k]
                    tag.attrib['exponent'] = str(e)
                    if m != 1:
                        tag.attrib['multiplier'] = str(m)
                        m = 1

            # Or... if the unit doesn't contain any of those seven, it must be
            # a dimensionless unit with a multiplier. These occur in CellML
            # definitions when unit mismatches are "resolved" by adding
            # conversion factors as units. This has no impact on the actual
            # equations...
            if m != 1:
                tag = et.SubElement(utag, 'unit')
                tag.attrib['units'] = si_units[myokit.units.dimensionless]
                tag.attrib['exponent'] = str(1)
                tag.attrib['multiplier'] = str(m)
                # m = 1

            # Add the new unit to the list
            unit_map[unit] = name

        # Sort by name
        sortkey = lambda x: x.qname()

        # Add variable and expression units
        for var in sorted(model.variables(deep=True), key=sortkey):
            add_unit(var.unit())
            for e in var.rhs().walk(myokit.Number):
                add_unit(e.unit())

        # Add si units to unit map
        for unit, name in si_units.items():
            unit_map[unit] = name

        # Add components
        # Components can correspond to Myokit components or variables with
        # children!
        ecomps = {}     # Components/Variables: elements (tags)
        cnames = {}     # Components/Variables: names (strings)
        unames = set()  # Unique name check

        def uname(name):
            # Create a unique component name
            i = 1
            r = name + '_'
            while name in unames:
                i += 1
                name = r + str(i)
            return name

        def export_nested_var(parent_tag, parent_name, var):
            # Create unique component name
            cname = uname(parent_name + '_' + var.uname())
            cnames[var] = cname
            unames.add(cname)

            # Create element
            ecomp = et.SubElement(emodel, 'component')
            ecomp.attrib['name'] = cname
            ecomps[var] = ecomp

            # Check for nested variables with children
            for kid in sorted(var.variables(), key=sortkey):
                if kid.has_variables():
                    export_nested_var(ecomp, cname, kid)

        for comp in sorted(model.components(), key=sortkey):
            # Create unique name
            cname = uname(comp.name())
            cnames[comp] = cname
            unames.add(cname)

            # Create element
            ecomp = et.SubElement(emodel, 'component')
            ecomp.attrib['name'] = cname
            ecomps[comp] = ecomp

            # Check for variables with children
            for var in sorted(comp.variables(), key=sortkey):
                if var.has_variables():
                    export_nested_var(ecomp, cname, var)

        # Collect oxmeta annotated variables
        oxmeta_vars = {}

        # Add variables
        def add_variable(eparent, var):
            evar = et.SubElement(eparent, 'variable')
            evars[var] = evar
            evar.attrib['name'] = var.uname()

            # Weblab oxmeta id given? Then add a cmeta id to reference via RDF
            # later.
            if 'oxmeta' in var.meta:
                # Ensure cmeta namespace is defined
                if 'xmlns:cmeta' not in emodel.attrib:
                    emodel.attrib['xmlns:cmeta'] = NS_CMETA

                # Add cmeta:id to variable
                cmeta_id = var.uname()
                evar.attrib['cmeta:id'] = cmeta_id

                # Store cmeta id and annotation for later
                oxmeta_vars[cmeta_id] = var.meta['oxmeta']

            # Add units
            unit = var.unit()
            unit = unit_map[unit] if unit else 'dimensionless'
            evar.attrib['units'] = unit

            # Add initial value
            init = None
            if var.is_literal():
                init = var.rhs().eval()
            elif var.is_state():
                init = var.state_value()
            if init is not None:
                evar.attrib['initial_value'] = myokit.strfloat(init)

        evars = {}
        for parent in sorted(ecomps, key=sortkey):
            eparent = ecomps[parent]
            for var in sorted(parent.variables(), key=sortkey):
                add_variable(eparent, var)

        # Add variable interfaces, connections
        deps = model.map_shallow_dependencies(
            omit_states=False, omit_constants=False)
        for var in sorted(evars, key=sortkey):
            evar = evars[var]
            # Scan all variables, iterate over the vars they depend on
            par = var.parent()
            lhs = var.lhs()
            dps = set(deps[lhs])

            # Add dependency on time for state variables
            if var.is_state():
                dps.add(time.lhs())

            for dls in sorted(dps, key=lambda x: x.var().qname()):
                dep = dls.var()
                dpa = dep.parent()
                # Parent mismatch: requires connection
                if par != dpa:
                    # Check if variable tag is present
                    epar = ecomps[par]
                    tag = epar.find('variable[@name="' + dep.uname() + '"]')
                    if tag is None:
                        # Create variable tag
                        tag = et.SubElement(epar, 'variable')
                        tag.attrib['name'] = dep.uname()

                        # Add unit
                        unit = dep.unit()
                        unit = unit_map[unit] if unit else 'dimensionless'
                        tag.attrib['units'] = unit

                        # Set interfaces
                        tag.attrib['public_interface'] = 'in'
                        edpa = ecomps[dpa]
                        tag = edpa.find(
                            'variable[@name="' + dep.uname() + '"]')
                        tag.attrib['public_interface'] = 'out'

                        # Add connection for this variable
                        comp1 = cnames[par]
                        comp2 = cnames[dpa]
                        vname = dep.uname()

                        # Sort components in connection alphabetically to
                        # ensure uniqueness
                        if comp2 < comp1:
                            comp1, comp2 = comp2, comp1

                        # Find or create connection
                        ctag = None
                        for con in emodel.findall('connection'):
                            ctag = con.find(
                                'map_components[@component_1="'
                                + comp1 + '"][@component_2="' + comp2 + '"]')
                            if ctag is not None:
                                break
                        if ctag is None:
                            con = et.SubElement(emodel, 'connection')
                            ctag = et.SubElement(con, 'map_components')
                            ctag.attrib['component_1'] = comp1
                            ctag.attrib['component_2'] = comp2
                        vtag = con.find(
                            'map_variables[@variable_1="' + vname
                            + '"][variable_2="' + vname + '"]')
                        if vtag is None:
                            vtag = et.SubElement(con, 'map_variables')
                            vtag.attrib['variable_1'] = vname
                            vtag.attrib['variable_2'] = vname

        # Add RDF for oxmeta annotated variables
        if oxmeta_vars:
            erdf = et.SubElement(emodel, 'rdf:RDF', {
                'xmlns:bqbiol': NS_BQBIOL,
                'xmlns:oxmeta': NS_OXMETA,
                'xmlns:rdf': NS_RDF,
                #'xmlns:rdfs': NS_RDFS,
            })
            for cmeta_id in sorted(oxmeta_vars):
                annotation = oxmeta_vars[cmeta_id]
                edesc = et.SubElement(erdf, 'rdf:Description')
                edesc.attrib['rdf:about'] = '#' + cmeta_id
                eis = et.SubElement(edesc, 'bqbiol:is')
                eis.attrib['rdf:resource'] = NS_OXMETA + quote(annotation)

        # Create CellMLWriter
        writer = cellml.CellMLExpressionWriter(units=unit_map)
        writer.set_element_tree_class(et)
        writer.set_time_variable(time)

        # Add equations
        def add_child_equations(parent):
            # Add the equations to a cellml component
            try:
                ecomp = ecomps[parent]
            except KeyError:
                return
            maths = et.SubElement(ecomp, 'math')
            maths.attrib['xmlns'] = NS_MATHML
            for var in sorted(parent.variables(), key=sortkey):
                # No equation for time variable
                if var == time:
                    continue
                # Literals already have an initial value
                if var.is_literal():
                    continue
                writer.eq(var.eq(), maths)
                add_child_equations(var)

        for comp in model.components():
            add_child_equations(comp)

        # Write xml to file
        doc = et.ElementTree(emodel)
        doc.write(path, encoding='utf-8', method='xml')
        if pretty_xml:
            # Create pretty XML
            import xml.dom.minidom as m
            xml = m.parse(path)
            with open(path, 'wb') as f:
                f.write(xml.toprettyxml(encoding='utf-8'))

        # Log any generated warnings
        self.logger().log_warnings()

    def supports_model(self):
        """
        Returns ``True``.
        """
        return True


si_units = {
    myokit.units.dimensionless: 'dimensionless',
    myokit.units.A: 'ampere',
    myokit.units.F: 'farad',
    myokit.units.kat: 'katal',
    myokit.units.lux: 'lux',
    myokit.units.Pa: 'pascal',
    myokit.units.T: 'tesla',
    myokit.units.Bq: 'becquerel',
    myokit.units.g: 'gram',
    myokit.units.K: 'kelvin',
    myokit.units.m: 'meter',
    myokit.units.V: 'volt',
    myokit.units.cd: 'candela',
    myokit.units.Gy: 'gray',
    myokit.units.kg: 'kilogram',
    myokit.units.m: 'metre',
    myokit.units.s: 'second',
    myokit.units.W: 'watt',
    myokit.units.C: 'celsius',
    myokit.units.H: 'henry',
    myokit.units.L: 'liter',
    myokit.units.mol: 'mole',
    #myokit.units.rad: 'radian',   # this overwrites dimensionless
}


si_exponents = {
    -24: 'yocto',
    -21: 'zepto',
    -18: 'atto',
    -15: 'femto',
    -12: 'pico',
    -9: 'nano',
    -6: 'micro',
    -3: 'milli',
    -2: 'centi',
    -1: 'deci',
    1: 'deka',
    2: 'hecto',
    3: 'kilo',
    6: 'mega',
    9: 'giga',
    12: 'tera',
    15: 'peta',
    18: 'exa',
    21: 'zetta',
    24: 'yotta',
}




#
# Imports a model definition from a CellML file
# Only partial CellML support is provided.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import os
import textwrap
import xml.dom.minidom
from collections import OrderedDict

import myokit
import myokit.units
from myokit.mxml import dom_child, dom_next
from myokit.formats.mathml import parse_mathml_rhs


# Strings in Python 2 and 3
try:
    basestring
except NameError:   # pragma: no cover
    basestring = str


class CellMLError(myokit.ImportError):
    """
    Raised when a fatal error occurs when importing CellML
    """


class CellMLImporter(myokit.formats.Importer):
    """
    This:class:`Importer <myokit.formats.Importer>` imports a model definition
    from CellML.
    """
    def __init__(self, verbose=False):
        super(CellMLImporter, self).__init__()
        self._generated_names = None

    def _flatten(self, node):
        """
        Reduces a node's contents to flat text and returns it.
        """
        def text(node, buff=None):
            if buff is None:
                buff = []
            if node.nodeType == node.TEXT_NODE:
                t = node.nodeValue
                t = t.encode('ascii', errors='replace').decode()
                t = t.replace('\n', ' ')
                t = t.replace('\r', ' ')
                t = t.replace('\f', ' ')
                t = t.strip()
                if t != '':
                    buff.append(t)
            else:
                for kid in node.childNodes:
                    text(kid, buff)
            return buff

        return textwrap.fill(
            '\n'.join(text(node)), 75, replace_whitespace=False)

    def info(self):
        """
        Returns a string containing information about this importer.
        """
        return "Loads a Model definition from a CellML file."

    def model(self, path):
        """
        Reads a CellML file and returns a:class:`myokit.Model`.
        """
        log = self.logger()
        log.clear()
        log.clear_warnings()

        # Reset list of generated names
        self._generated_names = {}

        # Parse xml file to dom
        path = os.path.abspath(os.path.expanduser(path))
        dom = xml.dom.minidom.parse(path)

        # Parse dom to model
        model = self._parse_model(dom)

        # Run model validation, order variables etc
        try:
            model.validate()
        except myokit.IntegrityError as e:
            log.warn('Integrity error found in model: ' + str(e))
        except Exception:  # pragma: no cover
            # Cover pragma: This really shouldn't happen!
            import traceback
            log.warn(
                'Exception occurred when validating model. '
                + traceback.format_exc())
        return model

    def _parse_model(self, dom):
        """
        Parses a dom tree into a myokit model.
        """
        model_tag = dom.getElementsByTagName('model')[0]
        name = self._sanitise_name(model_tag.getAttribute('name'))
        model = myokit.Model(name)
        model.meta['author'] = 'Cellml converter'

        # Parse model meta information
        desc = []
        for tag in model_tag.getElementsByTagName('documentation'):
            desc.append(self._flatten(tag).strip())
        desc = ('\n'.join(desc)).strip()
        if desc:
            model.meta['desc'] = desc

        # Check for unsupported CellML features
        # Check for <import> (allowed in CellML 1.1+)
        if model_tag.getElementsByTagName('import'):
            raise CellMLError('The CellML <import> tag is not supported.')

        # Check for <reaction> (allowed in CellML 1.0 and 1.1)
        if model_tag.getElementsByTagName('reaction'):
            raise CellMLError('The CellML <reaction> tag is not supported.')

        # Check for <factorial> (allowed in all CellML, but why?)
        if model_tag.getElementsByTagName('factorial'):
            self.logger().warn('The <factorial> tag is not supported.')

        # Check for MathML features not currently allowed in CellML
        # Check for <partialdiff> (not allowed in any CellML)
        if model_tag.getElementsByTagName('partialdiff'):
            self.logger().warn('The <partialdiff> tag is not supported.')

        # Check for <sum> (not allowed in any CellML)
        if model_tag.getElementsByTagName('sum'):
            self.logger().warn('The <sum> tag is not supported.')

        # Parse unit definitions
        si_units, munits, cunits = self._parse_units(model_tag)

        def convert_unit(unit):
            """
            Resolves a CellML unit (string) and returns a:class:`myokit.Unit`
            or ``None`` if successful, returns a string if conversion failed.
            """
            if unit:
                if str(unit) in si_units:
                    unit = si_units[unit]
                elif unit in munits:
                    unit = munits[unit]
                elif unit in cunits[cname]:
                    unit = cunits[cname][unit]
                else:
                    unit = str(unit)
            return unit

        # Parse components
        components = {}     # Dict of (component name, component) pairs
        for tag in model_tag.getElementsByTagName('component'):
            cname = tag.getAttribute('name')
            name = self._sanitise_name(cname)
            self.logger().log('Parsing component: ' + name)
            comp = model.add_component(name)
            components[cname] = comp

        # Parse group relationships
        # There are a number of different group types. Myokit handles the
        # "encapsulation" type of grouping, the rest can be ignored without
        # introducing errors.
        # Dict of encapsulation relations (component, parent component) pairs
        parents = {}

        def scan_encapsulated_children(parent, pcomp):
            """
            Reads parent/child relationships from a <group> or <component_ref>
            tag and adds them to the dict ``parents``.

            Argument ``parent`` should be a <component_ref> tag and ``pcomp``
            should be the corresponding cellml component object.
            """
            kid = dom_child(parent, 'component_ref')
            while kid is not None:
                # Get cellml component from name
                try:
                    comp = components[kid.getAttribute('component')]
                except KeyError:
                    raise CellMLError(
                        'Group registered for unknown component: '
                        + kid.getAttribute('component'))

                # Log relationship
                self.logger().log(
                    'Component <' + comp.qname() + '> is encapsulated in <'
                    + pcomp.qname() + '>.')

                # Add relationship
                parents[comp] = pcomp

                # Scan kid for children
                scan_encapsulated_children(kid, comp)

                # Move to next kid
                kid = dom_next(kid, 'component_ref')

        for group in model_tag.getElementsByTagName('group'):

            # Filter out encapsulation groups
            is_encapsulation = False
            for ref in group.getElementsByTagName('relationship_ref'):
                if ref.getAttribute('relationship') == 'encapsulation':
                    is_encapsulation = True
                    break
            if not is_encapsulation:
                continue

            # Parse and store relationships
            parent = dom_child(group, 'component_ref')
            while parent is not None:
                # Get cellml component from name
                try:
                    pcomp = components[parent.getAttribute('component')]
                except KeyError:
                    raise CellMLError(
                        'Group registered for unknown component: '
                        + parent.getAttribute('component'))
                # Add kids
                scan_encapsulated_children(parent, pcomp)
                # Search for next parent
                parent = dom_next(parent, 'component_ref')

        # Parse variables
        references = {}  # Dict (component name, (var name, var))
        interfaces = {}  # Dict (component name, (var name,(pub, pri, unit)))
        variables = {}   # Dict (component name, (var name, variable))
        values = {}      # Dict (component name, (var name, variable value))

        for ctag in model_tag.getElementsByTagName('component'):
            cname = ctag.getAttribute('name')
            comp = components[cname]
            references[cname] = rfs = {}
            interfaces[cname] = ifs = {}
            variables[cname] = vrs = {}
            values[cname] = vls = {}

            for vtag in ctag.getElementsByTagName('variable'):
                vname = vtag.getAttribute('name')

                # Get public and private interface
                pub = vtag.getAttribute('public_interface')
                pri = vtag.getAttribute('private_interface')
                if pub not in ('in', 'out'):
                    pub = None
                if pri not in ('in', 'out'):
                    pri = None

                # Get unit
                unit = convert_unit(vtag.getAttribute('units'))

                # Native variable? Then create
                if not (pub == 'in' or pri == 'in'):
                    name = self._sanitise_name(vname)
                    self.logger().log('Parsing variable: ' + name)
                    var = comp.add_variable(name)
                    vrs[vname] = var
                    init = str(vtag.getAttribute('initial_value'))
                    if init != '':
                        vls[vname] = init

                    # Set unit
                    if isinstance(unit, basestring):
                        var.meta['cellml_unit'] = unit
                    else:
                        var.set_unit(unit)

                    # Add resolved reference
                    rfs[vname] = var

                else:
                    # Otherwise, store as unresolved reference
                    rfs[vname] = None

                # Store reference information
                ifs[vname] = (pub, pri, unit)

        # Parse connections
        for tag in model_tag.getElementsByTagName('connection'):

            # Find linked components
            map_components = tag.getElementsByTagName('map_components')[0]
            cname1 = map_components.getAttribute('component_1')
            cname2 = map_components.getAttribute('component_2')
            for comp in (cname1, cname2):
                if comp not in components:
                    raise CellMLError(
                        'Connection found for unlisted component: <' + comp
                        + '>.')
            comp1 = components[cname1]
            comp2 = components[cname2]

            # If component is encapsulated, find parent
            try:
                par1 = parents[comp1]
            except KeyError:
                par1 = None
            try:
                par2 = parents[comp2]
            except KeyError:
                par2 = None

            # Get relevant lists for components
            ifs1 = interfaces[cname1]
            ifs2 = interfaces[cname2]
            rfs1 = references[cname1]
            rfs2 = references[cname2]

            # Find all references
            for pair in tag.getElementsByTagName('map_variables'):
                ref1 = pair.getAttribute('variable_1')
                ref2 = pair.getAttribute('variable_2')

                # Check interfaces
                try:
                    int1 = ifs1[ref1]
                except KeyError:
                    self.logger().warn(
                        'No interface found for variable <' + str(ref1)
                        + '>, unable to resolve connection.')
                    break
                try:
                    int2 = ifs2[ref2]
                except KeyError:
                    self.logger().warn(
                        'No interface found for variable <' + str(ref2)
                        + '>, unable to resolve connection.')
                    break

                # Determine direction of reference
                ref_to_one = None
                if int2[0] == 'in' and (par1 == par2 or par2 == comp1):
                    # Reference from comp2 to its parent or sibling comp1
                    ref_to_one = True
                elif int1[0] == 'in' and (par1 == par2 or par1 == comp2):
                    # Reference from comp1 to its parent or sibling comp2
                    ref_to_one = False
                elif int2[1] == 'in' and par1 == comp2:
                    # Reference from comp2 to its child comp1
                    ref_to_one = True
                elif int1[1] == 'in' and par2 == comp1:
                    # Reference from comp1 to its child comp2
                    ref_to_one = False
                else:
                    self.logger().warn(
                        'Unable to resolve connection between <'
                        + str(ref1) + '> in ' + str(comp1) + '('
                        + str(int1[0]) + ', ' + str(int1[1]) + ') and <'
                        + str(ref2) + '> in ' + str(comp2) + '('
                        + str(int2[0]) + ', ' + str(int2[1]) + ').')
                    continue

                # Check units
                if int1[2] != int2[2]:
                    self.logger().warn(
                        'Unit mismatch between <' + str(ref1) + '> in '
                        + str(int1[2]) + ' and <' + str(ref2) + '> given in '
                        + str(int2[2]) + '.')

                # Now point reference at variable or reference in other comp
                if ref_to_one:
                    rfs2[ref2] = (cname1, ref1)
                    self.logger().log(
                        'Variable <' + str(ref2) + '> in <' + str(cname2)
                        + '> points at <' + str(ref1) + '> in <' + str(cname1)
                        + '>.')
                else:
                    rfs1[ref1] = (cname2, ref2)
                    self.logger().log(
                        'Variable <' + str(ref1) + '> in <' + str(cname1)
                        + '> points at <' + str(ref2) + '> in <' + str(cname2)
                        + '>.')

        # Check for references that are never connected
        for cname, rfs in references.items():
            for vname, ref in rfs.items():
                if ref is None:
                    self.logger().warn(
                        'Unresolved reference <' + str(vname) + '> in'
                        ' component <' + str(cname) + '>. Creating a dummy'
                        ' variable with this name.')
                    c = components[cname]
                    v = c.add_variable(vname)
                    v.set_rhs(0)
                    rfs[vname] = v

        # The references should now all point to either a variable or a
        # reference to another variable. In the next step, these are resolved.
        for cname, rfs in references.items():
            for vname, ref in rfs.items():
                if type(ref) == tuple:
                    while True:
                        ref = references[ref[0]][ref[1]]
                        if type(ref) != tuple:
                            break
                    rfs[vname] = ref

        # MathML number post-processor to extract unit
        def npp(node, number):
            unit = convert_unit(node.getAttribute('cellml:units'))
            if isinstance(unit, myokit.Unit):
                return myokit.Number(number.eval(), unit)
            else:
                return number

        # MathML derivative post-processor to check if we're only using time
        # derivatives
        global time
        time = None

        def dpp(lhs):
            var = lhs.var()
            global time
            if time is None:
                time = var
            elif time != var:
                raise CellMLError(
                    'Found derivatives to two different variables: <'
                    + str(time) + '> and <' + str(var) + '>.')

        # MathML expression parser
        def mathml(node, rfs):
            return parse_mathml_rhs(
                node,
                var_table=rfs,
                logger=self.logger(),
                number_post_processor=npp,
                derivative_post_processor=dpp)

        # Parse expressions
        for ctag in model_tag.getElementsByTagName('component'):
            cname = ctag.getAttribute('name')
            math = ctag.getElementsByTagName('math')
            vrs = variables[cname]
            vls = values[cname]
            rfs = references[cname]
            n = 0
            for m in math:
                tag = dom_child(m)
                while tag:
                    if tag.tagName != 'apply':
                        raise CellMLError(
                            'Unexpected tag in expression: <' + tag.tagName
                            + '>, expecting <apply>.')

                    # First child of tag should be <eq />
                    eq = dom_child(tag, 'eq')
                    if not eq:
                        raise CellMLError(
                            'Unexpected content in <math><apply>, expecting'
                            ' <eq> but got <' + cname + '>.')

                    # Get lhs and rhs tags
                    lhs_tag = dom_next(eq)
                    rhs_tag = dom_next(lhs_tag)

                    # Parse lhs
                    lhs = mathml(lhs_tag, rfs)
                    if not isinstance(lhs, myokit.LhsExpression):
                        raise CellMLError(
                            'Error parsing equation: Expecting <ci> or <apply>'
                            ' after <eq> in "' + cname + '", got <'
                            + str(lhs_tag.tagName) + '> instead. Differential'
                            ' algebraic equations are not supported).')

                    # Check variable
                    var = lhs.var()
                    if var not in vrs.values():
                        raise CellMLError(
                            'Error: Equation found for unknown variable <'
                            + str(var) + '>.')

                    # Check derivatives
                    if lhs.is_derivative():
                        # Get CellML variable name
                        vname = dom_child(lhs_tag, 'ci')
                        vname = vname.firstChild.data.strip()
                        # Promote variable
                        try:
                            i = float(vls[vname])
                            del(vls[vname])
                        except KeyError:
                            self.logger().warn(
                                'No initial value found for <' + var.qname()
                                + '>.')
                            i = 0
                        var.promote(i)

                    # Parse rhs
                    rhs = mathml(rhs_tag, rfs)
                    try:
                        rhs.validate()
                        var.set_rhs(rhs)
                    except myokit.IntegrityError:
                        self.logger().warn(
                            'Unable to resolve RHS for "' + str(var) + '": '
                            + str(rhs))

                    # Continue
                    tag = dom_next(tag)
                    n += 1

            self.logger().log(
                'Found ' + str(n) + ' equations in ' + cname + '.')

        # Use remaining initial values (can be used to set constants)
        for cname, vls in values.items():
            vrs = variables[cname]
            for vname, val in vls.items():
                vrs[vname].set_rhs(myokit.Number(val))

        # Bind time variable to engine time
        if time is not None:
            time.set_rhs(0)
            time.set_binding('time')

        # Check for variables with no rhs that are never referenced
        no_rhs = [v for v in model.variables(deep=True) if v.rhs() is None]
        no_rhs = set(no_rhs)
        for var in no_rhs:
            self.logger().warn(
                'No expression for variable <' + var.qname() + '> is'
                ' defined. This variable will be set to zero.')
            var.set_rhs(myokit.Number(0))
        return model

    def _parse_units(self, model_tag):
        """
        Parses all cellml units into myokit units.

        Returns a tuple (munits, cunits) where munits is a dict mapping cellml
        unit names to myokit unit objects (or None objects if a unit couldn't
        be parsed). The cunits part maps cellml component names to dicts of the
        same structure.
        """
        # <units> Can be placed inside <model>, <component> or <import>
        # for <model> and <import> the units are global.
        # The <import> tag is not supported by this importer.
        # A Units tag can set base_units="yes" to define a new base unit: this
        # is not supported.
        si_units = {
            'dimensionless': myokit.units.dimensionless,
            'ampere': myokit.units.A,
            'farad': myokit.units.F,
            'katal': myokit.units.kat,
            'lux': myokit.units.lux,
            'pascal': myokit.units.Pa,
            'tesla': myokit.units.T,
            'becquerel': myokit.units.Bq,
            'gram': myokit.units.g,
            'kelvin': myokit.units.K,
            'meter': myokit.units.m,
            'radian': myokit.units.rad,
            'volt': myokit.units.V,
            'candela': myokit.units.cd,
            'gray': myokit.units.Gy,
            'kilogram': myokit.units.kg,
            'metre': myokit.units.m,
            'second': myokit.units.s,
            'watt': myokit.units.W,
            'celsius': myokit.units.C,
            'henry': myokit.units.H,
            'liter': myokit.units.L,
            'mole': myokit.units.mol,
            'siemens': myokit.units.S,
            'weber': myokit.units.Wb,
            'coulomb': myokit.units.C,
            'hertz': myokit.units.Hz,
            'litre': myokit.units.L,
            'newton': myokit.units.N,
            'sievert': myokit.units.Sv,
            'joule': myokit.units.J,
            'lumen': myokit.units.lm,
            'ohm': myokit.units.R,
            'steradian': myokit.units.sr,
        }
        si_prefixes = {
            'yotta': 1e24,
            'zetta': 1e21,
            'exa': 1e18,
            'peta': 1e15,
            'tera': 1e12,
            'giga': 1e9,
            'mega': 1e6,
            'kilo': 1e3,
            'hecto': 1e2,
            'deka': 1e1,            # US Style, as defined in CellML 1.0
            'deca': 1e1,            # SI prefix
            'deci': 1e-1,
            'centi': 1e-2,
            'milli': 1e-3,
            'micro': 1e-6,
            'nano': 1e-9,
            'pico': 1e-12,
            'femto': 1e-15,
            'atto': 1e-18,
            'zepto': 1e-21,
            'yocto': 1e-24,
        }

        class Unit:
            def __init__(self, name):
                self.name = name
                self.parts = []

        class Part:
            def __init__(self, base):
                self.base = base
                self.prefix = None
                self.multiplier = None
                self.exponent = None

        def parse(tag):
            """
            Parses <units> tags into Unit objects.
            """
            import myokit.formats.cellml as cellml

            name = tag.getAttribute('name')
            self.logger().log('Parsing unit: ' + name)
            if not cellml.is_valid_identifier(name):
                raise CellMLError(
                    'Unit name is not a valid CellML identifier: '
                    + str(name))
            unit = Unit(name)
            for part in tag.getElementsByTagName('unit'):
                if part.hasAttribute('offset'):
                    self.logger().warn(
                        'The "offset" attribute for <unit> tags is not'
                        ' supported.')
                p = Part(part.getAttribute('units'))
                x = part.getAttribute('prefix')
                if x:
                    p.prefix = str(x)
                x = part.getAttribute('multiplier')
                if x != '':
                    try:
                        p.multiplier = float(x)
                    except (TypeError, ValueError):
                        self.logger().warn(
                            'Unable to parse multiplier for <unit> tag:'
                            + str(p.exponent))
                x = part.getAttribute('exponent')
                if x != '':
                    try:
                        p.exponent = float(x)
                    except (TypeError, ValueError):
                        self.logger().warn(
                            'Unable to parse exponent for <unit> tag:'
                            + str(p.exponent))
                unit.parts.append(p)
            return unit

        # Parse units in model
        munits = []
        tag = dom_child(model_tag, 'units')
        while tag:
            munits.append(parse(tag))
            tag = dom_next(tag, 'units')

        # Parse units in components
        cunits = {}
        for tag in model_tag.getElementsByTagName('component'):
            cunits[tag.getAttribute('name')] = units = []
            for unit in tag.getElementsByTagName('units'):
                units.append(parse(unit))

        # Method to order units
        def order(units):
            """
            Orders a list of (name, parts) tuples so that none of the parts
            refer to a unit defined later in the list. Returns an OrderedDict
            mapping names to (name, parts) tuples.
            """
            todo = units
            units = OrderedDict()

            # List units that can already be referenced at this point
            okay = set(si_units.keys())

            # Run through todo list
            while todo:
                done = []
                for unit in todo:
                    ok = True
                    for part in unit.parts:
                        if part.base not in okay:
                            ok = False
                            break
                    if ok:
                        done.append(unit)
                        okay.add(unit.name)
                for unit in done:
                    units[unit.name] = unit
                    todo.remove(unit)
                if len(done) == 0:
                    break

            # Unable to resolve all units?
            if todo:
                for unit in todo:
                    self.logger().warn(
                        'Unable to resolve unit: ' + str(unit.name))
                    units[unit.name] = unit
            return units

        # Order units (units can refer to each other in a DAG form)
        munits = order(munits)
        for name, units in cunits.items():
            cunits[name] = order(units)

        # Convert units
        def convert(obj, local_map=None):
            """
            Converts a Unit object to a myokit unit.
            """
            base = myokit.units.dimensionless
            for part in obj.parts:
                # Get simple unit
                unit = None
                if str(part.base) in si_units:
                    unit = si_units[part.base]
                elif part.base in munits:
                    unit = munits[part.base]
                elif local_map and part.base in local_map:
                    unit = local_map[part.base]
                if unit is None:
                    self.logger().warn('Unknown base unit: ' + str(part.base))
                    return None
                # Add prefix
                if part.prefix is not None:
                    try:
                        unit *= si_prefixes[part.prefix]
                    except KeyError:
                        try:
                            # String such as 1.5 will raise ValueError
                            # As will string with text
                            unit *= 10**int(str(part.prefix))
                        except ValueError:
                            self.logger().warn(
                                'Unknown prefix in unit specification: "'
                                + str(part.prefix) + '".')
                            return None
                # Exponent (prefix part is exponentiated, multiplier is not)
                if part.exponent is not None:
                    e = int(part.exponent)
                    if abs(e - part.exponent) > 1e-15:
                        self.logger().warn(
                            'Non-integer exponents in unit specification are'
                            ' not supported: ' + str(part.exponent))
                        return None
                    unit **= e
                # Multiplier
                if part.multiplier is not None:
                    unit *= part.multiplier
                # Multiply base unit with this one
                base *= unit
            self.logger().log(
                'Converted unit "' + obj.name + '" to ' + str(base))
            return base

        # Convert all units in <model>
        for name, obj in munits.items():
            munits[name] = convert(obj)

        # Convert all units in components
        for cname, units in cunits.items():
            for name, obj in units.items():
                units[name] = convert(obj, units)

        # Return unit maps
        return si_units, munits, cunits

    def _sanitise_name(self, name):
        """
        Tests if a name is a valid Myokit name. Adapts it if it isn't.
        """
        import myokit.formats.cellml as cellml

        # Test if valid CellML first, complain if it isn't (be strict!)
        if not cellml.is_valid_identifier(name):
            raise CellMLError(
                'Invalid name (not a CellML identifier): ' + str(name))

        # Convert to and from ascii to get rid of special characters
        name = name.encode('ascii', errors='replace').decode('ascii')
        try:
            myokit.check_name(name)
            return name
        except myokit.InvalidNameError as e:
            try:
                return self._generated_names[name]
            except KeyError:
                self.logger().warn('Invalid name: ' + str(e))
                clean = 'generated_name_' + str(1 + len(self._generated_names))
                self._generated_names[name] = clean
                return clean

    def supports_model(self):
        """
        Returns True.
        """
        return True



#
# SetupTools script for Myokit
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals
from setuptools import setup, find_packages


# Get version number
import os
import sys
sys.path.append(os.path.abspath('myokit'))
from _myokit_version import __version__ as version  # noqa
sys.path.pop()
del(os, sys)


# Load text for description and license
with open('README.md') as f:
    readme = f.read()


# Go!
setup(
    # Module name (lowercase)
    name='myokit',

    # Version
    version=version,

    description='A simple interface to cardiac cellular electrophysiology',

    long_description=readme,
    long_description_content_type='text/markdown',

    license='BSD 3-clause license',

    author='Michael Clerx',

    author_email='michael@myokit.org',

    url='http://myokit.org',

    # Packages to include
    packages=find_packages(include=('myokit', 'myokit.*')),

    # Include non-python files (via MANIFEST.in)
    include_package_data=True,

    # Register myokit as a shell script
    entry_points={
        'console_scripts': ['myokit = myokit.__main__:main']
    },

    # List of dependencies
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib>=1.5',
        'configparser',
        'setuptools',
        'sympy',            # Used in lib.markov
        # PyQT or PySide?
        # (PySide is pip installable, Travis can get PyQt from apt)
    ],

    # Optional extras
    extras_require={
        'docs': [
            'sphinx>=1.5, !=1.7.3',     # Doc generation
        ],
        'dev': [
            'flake8>=3',                # Style checking
        ],
        'optional': [
            'cma',                      # Used in lib.fit
        ],
        'gui': ['pyqt5', 'sip'],
        'pyqt': ['pyqt5', 'sip'],
        'pyside': ['pyside2'],
    },
)
